import pandas as pd
import re
import nltk
import numpy as np
import pickle
import scipy
from sklearn.pipeline import Pipeline
from sklearn.feature_extraction.text import CountVectorizer, TfidfVectorizer
from nltk.tokenize import TreebankWordTokenizer
from nltk.corpus import stopwords
from nltk.stem import WordNetLemmatizer
from nltk.corpus import wordnet
nltk.download('punkt')
nltk.download('wordnet')
nltk.download('averaged_perceptron_tagger')
nltk.download('stopwords')
nltk.download('omw-1.4')

# Classify an ngram as unigram (1), bigram (2), etc.
def get_ngram_length(ngram_text):
    return len(ngram_text.split(" "))


# Check if a phrase is in the preprocessed text; helper function to filter keywords_df
def is_feature_in_text(row):
    return (row['feature'] in row['preprocessed_text']) and all(
        [word in row['preprocessed_text'].split() for word in row['feature'].split()])


# Converts NLTK POS (part-of-speech) tag to a Wordnet POS tag
def get_wordnet_pos(nltk_pos):
    if nltk_pos.startswith('J'):
        return wordnet.ADJ
    elif nltk_pos.startswith('V'):
        return wordnet.VERB
    elif nltk_pos.startswith('N'):
        return wordnet.NOUN
    elif nltk_pos.startswith('R'):
        return wordnet.ADV
    else:
        return wordnet.NOUN  # or None

def remove_stopwords_fun(df, stopwords_list=None):
    if 'stopwords_list' and isinstance(stopwords_list, pd.Series) and stopwords_list.dtype == 'object':
        stop_words = stopwords_list
    else:
        lowercase_alphabet = list(map(chr, range(97, 123)))
        stop_words = stopwords.words('english') + lowercase_alphabet
    stop_words = {word.lower() for word in stop_words}

    df = df.apply(lambda text: " ".join([word for word in nltk.word_tokenize(text) if word not in stop_words]))
    return df

def preprocess_text(input_df, text_column_name, id_column_name=None, text_normalization='lemmatization',
                    case_sensitive=False, remove_numbers=True, remove_special_characters=True, remove_stopwords=True):
    """
        Preprocess text

        This function returns the preprocessed text column along with ID column

        Parameters
        ----------
        input_df : data frame
            A data frame with a text column and ID column
        id_column_name : str
            ID column name in input_df
        text_column_name : str
            text column name in input_df
        text_normalization : str
            Refers to text normalization methods; either 'stemming' or 'lemmatization'. Any other value
            performs no normalization.
        case_sensitive: bool
            Whether upper and lowercase text should be treated differently
        remove_numbers: bool
            Remove digits from text
        remove_special_characters: bool
            remove nonalphanumerics
        remove_stopwords: bool
            remove default NLTK English stopwords

        Returns
        -------
        features_df: data frame
            preprocessed text column as data frame
        id_column: column
            corresponding id column to features_df
    """

    # Check if an ID column is given otherwise assign a default one
    if not (id_column_name and id_column_name in input_df.columns and input_df[id_column_name].nunique() == len(input_df)):
        id_column_name = "ID"
        input_df[id_column_name] = range(0, len(input_df))
    assert (input_df[id_column_name].nunique() == len(input_df))
    assert (id_column_name in input_df.columns and text_column_name in input_df.columns)

    # Grab the text column and ID column from the input data
    features_df = input_df[text_column_name]
    assert (isinstance(features_df, pd.Series))
    id_column = input_df[id_column_name]
    assert (isinstance(id_column, pd.Series))
    assert (len(features_df) == len(id_column))

    ### Preprocessing ###
    # Replace missing values with empty string
    features_df.fillna('', inplace=True)

    # Lowercase all text
    if not case_sensitive:
        features_df = features_df.apply(lambda text: text.lower())

    # Remove all numbers (standalone and within words)
    if remove_numbers:
        features_df = features_df.apply(lambda text: re.sub(r'[\d]', "", text))

    # Replace everything besides non-alphanumerics and whitespaces with empty string (meant to get rid of special characters, punctuation)
    if remove_special_characters:
        features_df = features_df.apply(lambda text: re.sub(r'[^a-zA-Z0-9\s]', " ", text))

    # Removes stopwords if flag is set to true
    if remove_stopwords:
        features_df = remove_stopwords_fun(features_df)

    # Perform either stemming or lemmatization or neither
    if text_normalization:
        if text_normalization == "stemming":
            stemmer = nltk.stem.SnowballStemmer('english')
            features_df = features_df.apply(
                lambda text: ' '.join([stemmer.stem(word) for word in nltk.word_tokenize(text)]))
        elif text_normalization == "lemmatization":
            lemmatizer = WordNetLemmatizer()
            features_df = features_df.apply(lambda text: " ".join(lemmatizer.lemmatize(word, get_wordnet_pos(nltk_pos))
                                                                  for word, nltk_pos in
                                                                  nltk.pos_tag(nltk.word_tokenize(text))))
        else:
            pass

    # store tokenized text (list of str lists) as Binary in document property
    tokenize_text = [[tok for tok in text.split()] for text in features_df]
    tokenize_text = [elem for elem in tokenize_text if len(elem) > 0]
    store_tokenize_text = pickle.dumps(tokenize_text)  # Read: pickle.loads(<document property>)

    # Add transformed text column with original text columns; retains empty text rows
    preprocessed_text = input_df[text_column_name].to_frame()
    preprocessed_text.rename(columns={text_column_name: "original_text"}, inplace=True)
    preprocessed_text["preprocessed_text"] = features_df
    preprocessed_text['id'] = id_column

    # Remove empty text rows
    features_df = features_df[features_df != ""]
    which_indices = list((features_df != "").index)
    id_column = id_column.loc[which_indices]
    assert (list(id_column.index) == list(features_df.index))

    return features_df, id_column

def get_features(features_df, id_column, ngram_method='tfidf', ngram_size=3, max_features=1000, keywords_per_doc=5):
    """
        Extracts N-gram/Count Features

        This function returns tables referencing the N-gram or frequency matrices, top N-grams, and keywords

        Parameters
        ----------
        features_df : data frame
            A data frame with a text column and ID column
        id_column : str
            ID column name in input_df
        ngram_method : str
            either 'tfidf' for term frequency inverse document frequency or 'tf' for term frequency
        ngram_size : int
            max ngram sizes computed; lower bound is 1
        max_features: int
            max features in the ngram matrix. should be greater than 0
        keywords_per_doc: int
            max number of keywords per document to retrieve

        Returns
        -------
        Ngram_mat: scipy sparse matrix
            ngram_matrix often saved as binary document property
        top_ngrams_df: data frame
            top ngrams across corpus
        keywords_df:
            top keywords per document
        text_order_df: data frame
            order of text in ngram matrix
        ngrams_order_df: data frame
            order of ngrams in ngram matrix
    """

    Ngram_mat, top_ngrams_df, text_order_df, ngrams_order_df, keywords_df = None, None, None, None, None

    ### Ngram Operations ###
    try:
        # Calculate Ngram frequency matrix
        pipeline = Pipeline([('count_vectorizer', CountVectorizer(ngram_range=(1, ngram_size),
                                                                  tokenizer=TreebankWordTokenizer().tokenize))])
        Ngram_mat = pipeline.fit_transform(features_df)  # sparse array
        ngram_features = np.array(pipeline['count_vectorizer'].get_feature_names_out())

        # number of times each ngram appears across whole corpus
        ngram_frequencies = Ngram_mat.sum(axis=0)  # sparse matrix operation returns matrix
        ngram_frequencies = np.asarray(ngram_frequencies).flatten()

        # top features and frequencies are in order from most frequent to least frequent
        top_ngram_features = ngram_features[np.argsort(ngram_frequencies)[::-1]]
        top_ngram_frequencies = ngram_frequencies[np.argsort(ngram_frequencies)[::-1]]

        # retrieving most frequent features and optionally changing matrix to tfidf
        if ngram_method == 'tfidf':
            pipeline = pipeline = Pipeline([('tfidf_vectorizer', TfidfVectorizer(ngram_range=(1, ngram_size),
                                                                                 tokenizer=TreebankWordTokenizer().tokenize,
                                                                                 sublinear_tf=True))])
            Ngram_mat = pipeline.fit_transform(features_df)  # sparse array

        if not (max_features != None and max_features > 0 and max_features <= 10000):
            max_features = 1000

        # get subset of ngram matrix by most frequent ngrams (number set by max_features)
        subset_ngrams_mat_indices = np.argsort(ngram_frequencies)[::-1][:max_features]
        subset_ngrams_mat_features = top_ngram_features[:max_features]
        subset_ngrams_mat = Ngram_mat[:, subset_ngrams_mat_indices]

        # Save ngram matrix
        # scipy.sparse.save_npz('ngram_matrix.npz', subset_ngrams_mat)

        # format ngram frequencies into dataframe
        top_ngrams_df = pd.DataFrame()
        top_ngrams_df['ngram'] = top_ngram_features
        top_ngrams_df['frequencies'] = top_ngram_frequencies
        top_ngrams_df['length'] = top_ngrams_df['ngram'].apply(lambda text: get_ngram_length(text))

        # Order of text rows for subset_ngrams_mat
        text_order_df = pd.DataFrame()
        corpus = list(features_df)
        text_order_df['ngram_mat_row_number'] = [i for i in range(len(corpus))]

        text_order_df['id'] = list(id_column)
        text_order_df['preprocessed_text'] = corpus
        assert (subset_ngrams_mat.shape[0] == len(text_order_df))

        # Order of ngrams columns for subset_ngrams_mat
        ngrams_order_df = pd.DataFrame()
        features = [str(elem) for elem in list(subset_ngrams_mat_features)]
        ngrams_order_df['ngram_mat_column_number'] = [i for i in range(len(features))]
        ngrams_order_df['ngram'] = features
        assert (subset_ngrams_mat.shape[1] == len(ngrams_order_df))

        # store ngrams matrix as Binary in document property
        store_ngrams_mat = pickle.dumps(subset_ngrams_mat)  # Read: pickle.loads(<document property>)

        ### Top Scoring Words per Document ###
        dense_ngrams_mat = subset_ngrams_mat.toarray()
        top_keywords_indices = np.argpartition(dense_ngrams_mat, -keywords_per_doc, axis=1)[:, -keywords_per_doc:]
        all_features, all_scores, all_ids, all_texts = [], [], [], []
        for i in range(len(top_keywords_indices)):
            scores_per_doc = dense_ngrams_mat[i, top_keywords_indices[i, :]]
            sort_ind_per_doc = np.argsort(scores_per_doc)[::-1]
            scores_per_doc = scores_per_doc[sort_ind_per_doc]
            features_per_doc = subset_ngrams_mat_features[top_keywords_indices[i, :]][sort_ind_per_doc]
            id_per_doc = text_order_df.iloc[i]['id']
            text_per_doc = text_order_df.iloc[i]['preprocessed_text']

            all_features += [str(elem) for elem in features_per_doc]
            all_scores += [elem for elem in scores_per_doc]
            all_ids += [id_per_doc] * keywords_per_doc
            all_texts += [text_per_doc] * keywords_per_doc
        assert (len(all_features) == len(all_scores) == len(all_ids) == len(all_texts))
        keywords_df = pd.DataFrame.from_dict(
            data={'id': all_ids, 'preprocessed_text': all_texts, 'feature': all_features, 'score': all_scores})
        keywords_df['feature_in_text'] = keywords_df.apply(is_feature_in_text, axis=1)
        keywords_df = keywords_df[keywords_df['feature_in_text'] == True]
        keywords_df.drop(columns='feature_in_text', inplace=True)

    except Exception as err:
        print(err)

    return Ngram_mat, top_ngrams_df, text_order_df, ngrams_order_df, keywords_df

#example of sequential function calls
def main(input_df, text_column_name):
    features_df, id_column = preprocess_text(input_df, text_column_name)
    Ngram_mat, top_ngrams_df, text_order_df, ngrams_order_df, keywords_df = get_features(features_df, id_column)
    return Ngram_mat, top_ngrams_df, text_order_df, ngrams_order_df, keywords_df



# In standalone mode (running 'main'), create matplotlib plots as well if PLOTS = True.
# 'main' sets PLOTS = True, and resets it to FALSE.
# You can also set PLOT = True here when using the functions in Spotfire, in which
# case matplotlib plots will pop up, without you having to create Spotfire visualizations,
# which may be a good debugging tool to have.

PLOTS = False


# Ideally, we would make this function non-exportable. That, however, is not really an option in Python.
# Instead signal to other programmers using a leading underscore in the name.

def _ExtractColumnLists(dfX, max_low_cardinality, max_high_cardinality):
    
    '''
    Generates four lists that the ColumnTransformers need in the pipeline definition.
    The responsibility of setting default values for max_low_cardinality and max_high_cardinality is
    placed on the callers.

    input: dfX = dataframe
           max_low_cardinality = maximum cardinality for one hot encoding, otherwise target encoding
           max_high_cardinality = maximum cardinality, if higher the feature is ignored

    output: lists of numeric, categoric, low and high cardinality categoric features
    
    '''
    
    import pandas as pd
    import numpy as np
    
    categoric_columns = dfX.select_dtypes(exclude=np.number).columns
    numeric_columns = dfX.select_dtypes(include=np.number).columns

    low_cardinality_categoric_columns = []
    high_cardinality_categoric_columns = []
    for col in categoric_columns:
        non_nan_values = dfX.loc[:,col].dropna().values
        cardinality = len(np.unique(non_nan_values))
        if cardinality < max_low_cardinality + 1:
            low_cardinality_categoric_columns.append(col)
        elif cardinality < max_high_cardinality + 1:
            high_cardinality_categoric_columns.append(col)
                    
    return numeric_columns, categoric_columns, low_cardinality_categoric_columns, high_cardinality_categoric_columns


def _ExtractGridSearchData(cv_results_, grid_args, mode):

    '''
    Creates dataframe df_cv_results for plotting a parallel coordinates plot with grid search results
    
    mode: 'classifier' or 'regressor'
    
    '''

    import pandas as pd

    df_cv_results = pd.DataFrame.from_dict(cv_results_['params'])
    df_cv_results['mean_test_score'] = (-1 ** int(mode=='regression')) * cv_results_['mean_test_score']
    df_cv_results = df_cv_results.sort_values('mean_test_score', ascending=(mode=='regression'))

    if False:

        import matplotlib.pyplot as plt
        from yellowbrick.features import ParallelCoordinates
        
        df_cv_results_plot = df_cv_results.copy()
        df_cv_results_plot['mean_test_score_bin'] = pd.cut(df_cv_results_plot['mean_test_score'], bins=df_cv_results.shape[0], labels=False)
        df_cv_results_plot = df_cv_results_plot.select_dtypes(exclude='object')
        df_cv_results_plot.loc[:,'mean_test_score_bin'] = df_cv_results_plot.loc[:,'mean_test_score_bin']
        df_cv_results_plot = df_cv_results_plot.sort_values(by='mean_test_score_bin')
        
        features = df_cv_results_plot.columns[:-1]
        classes = list(sorted(df_cv_results_plot['mean_test_score_bin'].unique()))
        
        df_cv_results_plot.loc[:,'mean_test_score_bin'] = df_cv_results_plot.loc[:,'mean_test_score_bin'].astype(str)

        print(df_cv_results_plot[['mean_test_score','mean_test_score_bin']])
        print(df_cv_results_plot.dtypes)

        features = df_cv_results_plot.columns[:-1]
        classes = list(df_cv_results_plot['mean_test_score_bin'].unique())

        print(features)
        print(classes)
    
        fig, ax = plt.subplots()
        kwargs = {}
        pc = ParallelCoordinates(classes=classes, features=features, normalize='minmax', colormap='coolwarm', **kwargs)
        pc.fit(df_cv_results_plot[features], df_cv_results_plot['mean_test_score_bin'])
        pc.transform(df_cv_results_plot[features])
        ax.set_xticklabels(ax.get_xticklabels(), rotation=45)
        pc.poof()
        plt.show()
    

    return df_cv_results


def CreateClassificationPipeline(dfX,
                                 max_low_cardinality=6,
                                 max_high_cardinality=50,
                                 algo='randomforest',
                                 classifier_args={}):
    
    '''
    Creates a Pipeline object.
    
    Categorical pipeline:
    SimpleImputer + [OneHotEncoder (if low cardinality) | TargetEncoder (if higher cardinality) | drop feature (highest cardinality)]

    Numeric pipeline:
    SimpleImputer(mean) + StandardScaler + Addition new feature: outlier_score

    Classifier:
    RandomForestClassifier | XGBClassifier | LogisticRegression | KerasClassifier

    input: dfX = training data, sued here to extract num v low cardinality cat v high cardinality cat
           max_low_cardinality (optional) = max cardinality for one hot encoding, otherwise target encoding
           max_high_cardinality (optional) = features with higher cardinality are ignored
           algo (optional) = classification algorithm
           classifier_args (optional) = dictionary of arguments passed to the classifier specified in algo

    output: classifier pipeline object (untrained) 
    
    '''

    from sklearn.compose import ColumnTransformer
    from sklearn.impute import SimpleImputer
    from sklearn.pipeline import Pipeline
    from sklearn.preprocessing import StandardScaler, OneHotEncoder
    from sklearn.preprocessing import FunctionTransformer
    
    from category_encoders.target_encoder import TargetEncoder
    
    from sklearn.ensemble import RandomForestClassifier
    from sklearn.linear_model import LogisticRegression
    from xgboost import XGBClassifier

    import numpy as np

    if algo == 'randomforest':
        
        # possibly remove k,v pairs from input classifier_args if we want them hardcoded here
        classifier = RandomForestClassifier(**classifier_args)
        
    elif algo == 'xgboost':
        
        # remove k,v pairs from input classifier_args that we want hardcoded here
        if 'objective' in classifier_args.keys():
            del classifier_args['objective']
        if 'eval_metric' in classifier_args.keys():
            del classifier_args['eval_metric']
        if 'use_label_encoder' in classifier_args.keys():
            del classifier_args['use_label_encoder']
        classifier = XGBClassifier(objective='binary:logistic',
                                   eval_metric='logloss',
                                   use_label_encoder=False, 
                                   **classifier_args)
        
    elif algo == 'logisticregression':
        
        # possibly remove k,v pairs from classifier_args if we want them hardcoded here
        classifier = LogisticRegression(**classifier_args)
        
    elif algo == 'neuralnet':
        
        import tensorflow as tf
        from tensorflow import keras
        from scikeras.wrappers import KerasClassifier

        tf.random.set_seed(59)

        if 'patience' in classifier_args.keys():
            patience = classifier_args['patience']
        else:
            patience = 10

        if 'epochs' in classifier_args.keys():
            epochs = classifier_args['epochs']
        else:
            epochs = 100    

        if 'batch_size' in classifier_args.keys():
            batch_size = classifier_args['batch_size']
        else:
            batch_size = min(100, int(dfX.shape[0]/10))

        if 'hidden_layers_sizes' in classifier_args.keys():
            hidden_layers_sizes = classifier_args['hidden_layers_sizes']
        else:
            nn = int(np.sqrt(0.5 * dfX.shape[0]))
            hidden_layers_sizes = (nn,nn,)

        if 'learning_rate' in classifier_args.keys():
            learning_rate = classifier_args['learning_rate']
        else:
            learning_rate = 0.001

        if 'dropout' in classifier_args.keys():
            dropout = classifier_args['dropout']
        else:
            dropout = 0.1
            
        early_stopping_callback = tf.keras.callbacks.EarlyStopping(monitor='val_loss', 
                                                                   mode='min',
                                                                   patience=patience,
                                                                   restore_best_weights=True,
                                                                   verbose=1)

        # For now: disable one-hot encoding and dropping of non-numeric variables,
        # to avoid logic to determine input_shape inside the pipeline. That is:
        # we use targetEncoder for all non-numeric variables

        n_in = dfX.shape[1]
        max_low_cardinality = 0
        max_high_cardinality = 10000

        def build_classifier_neural_net(hidden_layers_sizes, dropout, n_in):

            model = keras.models.Sequential()
            model.add(keras.layers.Input(shape=(n_in,)))
            for hidden_layer_size in hidden_layers_sizes:
                model.add(keras.layers.Dense(hidden_layer_size, activation='relu'))
                model.add(keras.layers.Dropout(dropout))
            model.add(keras.layers.Dense(1, activation='sigmoid'))
            return model
  
        params_keras = {'model__n_in':n_in,
                        'model__hidden_layers_sizes':hidden_layers_sizes,
                        'model__dropout':dropout,
                        'optimizer':'adam',
                        'optimizer__learning_rate':learning_rate,
                        'validation_split':0.1,
                        'epochs':epochs,
                        'batch_size':batch_size,
                        'loss':'binary_crossentropy',
                        'callbacks':[early_stopping_callback],
                        'verbose':0
                        }
        
        classifier = KerasClassifier(model=build_classifier_neural_net, **params_keras) 
        
    else:
        
        raise ValueError('The algorithm {} is not supported. \
                          The options are: randomforest, xgboost, logisticregression, neuralnet'.format(algo))
    
    numeric_columns, categoric_columns, low_cardinality_categoric_columns, high_cardinality_categoric_columns = \
        _ExtractColumnLists(dfX, max_low_cardinality, max_high_cardinality)

    high_cardinality_cat_encoder_pipe = Pipeline([
        ('imputer', SimpleImputer(strategy='constant', fill_value='NA')),
        ('high_cardinality_encoding', TargetEncoder())
    ])

    low_cardinality_cat_encoder_pipe = Pipeline([
        ('imputer', SimpleImputer(strategy='constant', fill_value='NA')),
        ('low_cardinality_encoding', OneHotEncoder(handle_unknown='ignore'))
    ])

    cat_encoder_pipe = ColumnTransformer([
        ('high_cardinality', high_cardinality_cat_encoder_pipe, high_cardinality_categoric_columns),
        ('low_cardinality', low_cardinality_cat_encoder_pipe, low_cardinality_categoric_columns)    
    ])

    # categoric preprocessing
    categoric_pipe = Pipeline([
        ('encoder', cat_encoder_pipe)
    ])

    numeric_pipe = Pipeline([
        ('imputer', SimpleImputer()),
        ('scaler', StandardScaler())
    ])

    # this allows for separate processing of categoric and numeric features
    preprocessing = ColumnTransformer(
        [('cat', categoric_pipe, categoric_columns),
         ('num', numeric_pipe, numeric_columns)])

    # final random forest pipeline
    classifier_pipeline = Pipeline([
        ('preprocess', preprocessing),
        ('classifier', classifier)
    ])
    
    return classifier_pipeline



def TrainXGBClassifier(pipeline, dfX_train, dfy_train, dfX_test=None, dfy_test=None, classifier_args={}):

    """
    Dedicate train function for pipelines with an XGB classifier, supporting early_stopping_rounds

    """

    from sklearn.pipeline import Pipeline
    from xgboost import XGBClassifier

    if 'early_stopping_rounds' in classifier_args.keys() and dfX_test is not None and dfy_test is not None:

        fitted_preprocessor = pipeline['preprocess'].fit(dfX_train, dfy_train)
        dfX_test_trans = fitted_preprocessor.transform(dfX_test)
        fit_params = {'classifier__eval_set':[(dfX_test_trans, dfy_test)],
                      'classifier__verbose':True
                     }
    else:
        fit_params = {}

    model = pipeline.fit(dfX_train, dfy_train, **fit_params)

    return model


def TrainClassifier(pipeline, dfX_train, dfy_train, dfX_test=None, dfy_test=None, classifier_args={}):

    """
    Dedicate train function for classifier pipelines

    """

    from sklearn.pipeline import Pipeline

    algo = str(pipeline.named_steps['classifier']).split('(')[0]


    if algo != 'XGBClassifier':

        model = pipeline.fit(dfX_train, dfy_train)

    else:

        if 'early_stopping_rounds' in classifier_args.keys() and dfX_test is not None and dfy_test is not None:

            fitted_preprocessor = pipeline['preprocess'].fit(dfX_train, dfy_train)
            dfX_test_trans = fitted_preprocessor.transform(dfX_test)
            fit_params = {'classifier__eval_set':[(dfX_test_trans, dfy_test)],
                          'classifier__verbose':True
                         }
        else:
            fit_params = {}

        model = pipeline.fit(dfX_train, dfy_train, **fit_params)

    return model

    
    
def ClassificationGridSearchCV(pipeline,
                               dfX_train, dfy_train,
                               dfX_test=None, dfy_test=None,
                               cv=4,
                               scoring='roc_auc',
                               pos_label=1,
                               classifier_grid_args={}):
    
    '''
    Performs a grid search to train the pipeline 
    scoring: 'roc_auc' or 'average_precision'

    input: pipeline, pipeline to be trained using grid search
           dfX_train, dfy_train are dataframes
           dfX_test, dfy_test are dataframes only needed for xgb pipelines for early stopping (to do)
           scoring = 'roc_auc' or 'average_precision'

    output: trained classifier pipeline
    
    '''

    import numpy as np
    import pandas as pd
    from sklearn.model_selection import GridSearchCV
    
    algo = str(pipeline.named_steps['classifier']).split('(')[0]
 
    if algo == 'RandomForestClassifier':
        
        print('Grid search with RandomForestClassifier')
        
        if not classifier_grid_args:
            pipeline_params = {
                           'preprocess__num__imputer__strategy':['mean'],
                           'classifier__n_estimators':[100],
                           'classifier__min_samples_leaf':[1,2,3],
                           'classifier__class_weight':['balanced',None],
                           }
        else:
            pipeline_params = {
                           'preprocess__num__imputer__strategy':['mean'],
                           }

            for k,v in classifier_grid_args.items():
                pipeline_params['classifier__' + k] = v  
              
        fit_params = {         
                     }

    elif algo == 'XGBClassifier':
        
        print('Grid search with XGBClassifier')
        
        y_pos = len(dfy_train[dfy_train == pos_label])
        y_min = len(dfy_train[dfy_train != pos_label])
        scale_pos_weight = (1.0*y_min)/y_pos
        
        if not classifier_grid_args:
            
            pipeline_params = {
                           'classifier__n_estimators':[50,100],
                           'classifier__scale_pos_weight':[scale_pos_weight, 1],
                           'classifier__min_child_weight':[1,2],
                           'classifier__learning_rate':[0.1,0.03],
                           'classifier__max_depth':[5,6],
                           'classifier__use_label_encoder':[False],
                           'classifier__objective':['binary:logistic'],
                           }
        else:
            
            pipeline_params = {
                           'preprocess__num__imputer__strategy':['mean'],
                           }

            for k,v in classifier_grid_args.items():
                           pipeline_params['classifier__' + k] = v   
        fit_params = {         
                     }
        
    elif algo == 'LogisticRegression':
        
        print('Grid search with LogisticRegression')
        
        if not classifier_grid_args:
            
            pipeline_params = {
                           'preprocess__num__imputer__strategy':['mean'],
                           'classifier__class_weight':['balanced', None],
                           'classifier__C':[0.01, 0.1, 1, 10, 100],
                          }
        else:
            
            pipeline_params = {
                           'preprocess__num__imputer__strategy':['mean'],
                           }
            
            for k,v in classifier_grid_args.items():
                           pipeline_params['classifier__' + k] = v              
        fit_params = {  
                     }

    elif algo == 'KerasClassifier':

        print('Grid search with KerasClassifier')
        
        if not classifier_grid_args:
            
            param_grid = {'classifier__model__n_in':[dfX_train.shape[1]],
                          'classifier__model__hidden_layers_sizes':[(1000,),(500,500,),(333,334,333,)],
                          'classifier__optimizer__learning_rate':[0.003,0.01],
                          'classifier__batch_size':[100,1000],
                          'classifier__epochs':[200]
                         }

        else:

            param_grid = {'classifier__model__n_in':[dfX_train.shape[1]]
                         }

            for k,v in classifier_grid_args.items():

                if k in ('hidden_layers_sizes','dropout'):
                    param_grid['classifier__model__' + k] = v
                elif k in ('learning_rate'):
                    param_grid['classifier__optimizer__' + k] = v
                else:
                    param_grid['classifier__' + k] = v
        
    
    if algo != 'KerasClassifier':
        
        grid = GridSearchCV(estimator=pipeline, 
                            param_grid=pipeline_params, 
                            cv=cv,
                            scoring=scoring, 
                            refit=True,
                            verbose=1,
                            n_jobs=1)
    else:

        grid = GridSearchCV(pipeline, param_grid=param_grid, scoring='roc_auc')

    import time
    start = time.time()
    trained_classifier = grid.fit(dfX_train, dfy_train) #, **fit_params)
    print("Running time =", time.time() - start, "seconds")

    df_grid_search = _ExtractGridSearchData(grid.cv_results_, classifier_grid_args, 'classification')
        
    return trained_classifier



def CreateRegressionPipeline(dfX,
                             max_low_cardinality=6,
                             max_high_cardinality=50,
                             algo='randomforest',
                             regressor_args={}):

    '''
    Creates a Pipeline object
    Categorical pipeline: SimpleImputer + OneHotEncoder (low cardinailty) + TargetEncoder (high cardinality)
    Numeric pipeline: SimpleImputer + StandardScaler
    Regressor: RandomForestRegressor | XGBRegressor | Ridge | Lasso | KerasRegressor

    input: dfX = training data, sued here to extract num v low cardinality cat v high cardinality cat
           max_low_cardinality = max cardinality for one hot encoding, otherwise target encoding
           max_high_cardinality = features with higher cardinality are ignored
           algo = classification algorithm

    output: regressor pipeline object (untrained)
    
    '''

    from sklearn.compose import ColumnTransformer
    from sklearn.impute import SimpleImputer
    from sklearn.pipeline import Pipeline
    from sklearn.preprocessing import StandardScaler, OneHotEncoder
    
    from category_encoders.target_encoder import TargetEncoder
    
    from sklearn.ensemble import RandomForestRegressor
    from sklearn.linear_model import Ridge, Lasso
    from xgboost import XGBRegressor

    import numpy as np

    if algo == 'randomforest':
        
        regressor = RandomForestRegressor(**regressor_args)
        
    elif algo == 'xgboost':
        
        # remove k,v pairs from regressor_args that we want them hardcoded here
        if 'objective' in regressor_args.keys():
            del regressor_args['objective']
        if 'eval_metric' in regressor_args.keys():
            del regressor_args['eval_metric']
        if 'use_label_encoder' in regressor_args.keys():
            del regressor_args['use_label_encoder']
        regressor = XGBRegressor(objective='reg:squarederror',
                                 use_label_encoder=False,
                                 **regressor_args)
        
    elif algo == 'ridge':
        
        regressor = Ridge(**regressor_args)
        
    elif algo == 'lasso':
        
        regressor = Lasso(**regressor_args)

    elif algo == 'neuralnet':
        
        import tensorflow as tf
        from tensorflow import keras
        from scikeras.wrappers import KerasRegressor

        tf.random.set_seed(59)

        if 'patience' in regressor_args.keys():
            patience = regressor_args['patience']
        else:
            patience = 10

        if 'epochs' in regressor_args.keys():
            epochs = regressor_args['epochs']
        else:
            epochs = 100    

        if 'batch_size' in regressor_args.keys():
            batch_size = regressor_args['batch_size']
        else:
            batch_size = min(100, int(dfX.shape[0]/10))

        if 'hidden_layers_sizes' in regressor_args.keys():
            hidden_layers_sizes = regressor_args['hidden_layers_sizes']
        else:
            nn = int(np.sqrt(0.5 * dfX.shape[0]))
            hidden_layers_sizes = (nn,nn,)

        if 'learning_rate' in regressor_args.keys():
            learning_rate = regressor_args['learning_rate']
        else:
            learning_rate = 0.001

        if 'dropout' in regressor_args.keys():
            dropout = regressor_args['dropout']
        else:
            dropout = 0.1
            
        early_stopping_callback = tf.keras.callbacks.EarlyStopping(monitor='val_loss', 
                                                                   mode='min',
                                                                   patience=patience,
                                                                   restore_best_weights=True,
                                                                   verbose=1)

        # For now: disable one-hot encoding and dropping of non-numeric variables,
        # to avoid logic to determine input_shape inside the pipeline. That is:
        # we use targetEncoder for all non-numeric variables
        
        n_in = dfX.shape[1]
        max_low_cardinality = 0
        max_high_cardinality = 10000

        def build_regressor_neural_net(hidden_layers_sizes, dropout, n_in):

            model = keras.models.Sequential()
            model.add(keras.layers.Input(shape=(n_in,)))
            for hidden_layer_size in hidden_layers_sizes:
                model.add(keras.layers.Dense(hidden_layer_size, activation='relu'))
                model.add(keras.layers.Dropout(dropout))
            model.add(keras.layers.Dense(1, activation=None))
            return model
  
        params_keras = {'model__n_in':n_in,
                        'model__hidden_layers_sizes':hidden_layers_sizes,
                        'model__dropout':dropout,
                        'optimizer':'adam',
                        'optimizer__learning_rate':learning_rate,
                        'validation_split':0.1,
                        'epochs':epochs,
                        'batch_size':batch_size,
                        'loss':'mean_squared_error',
                        'callbacks':[early_stopping_callback],
                        'verbose':0
                        }
        
        regressor = KerasRegressor(model=build_regressor_neural_net, **params_keras)       
        
    else:
        
        raise ValueError('The algorithm {} is not supported. The options are: randomforest, xgboost, ridge, neuralnet.'.format(algo))
    
    numeric_columns, categoric_columns, low_cardinality_categoric_columns, high_cardinality_categoric_columns = \
        _ExtractColumnLists(dfX, max_low_cardinality, max_high_cardinality)

    high_cardinality_cat_encoder_pipe = Pipeline([
        ('imputer', SimpleImputer(strategy='constant', fill_value='NA')),
        ('high_cardinality_encoding', TargetEncoder()),
        ('high_cardinality_scaling', StandardScaler())
    ])

    low_cardinality_cat_encoder_pipe = Pipeline([
        ('imputer', SimpleImputer(strategy='constant', fill_value='NA')),
        ('low_cardinality_encoding', OneHotEncoder(handle_unknown='ignore'))
    ])

    cat_encoder_pipe = ColumnTransformer([
        ('high_cardinality', high_cardinality_cat_encoder_pipe, high_cardinality_categoric_columns),
        ('low_cardinality', low_cardinality_cat_encoder_pipe, low_cardinality_categoric_columns)    
    ])

    # categoric preprocessing
    categoric_pipe = Pipeline([
        #('imputer', SimpleImputer(strategy='constant', fill_value='NA')),
        ('encoder', cat_encoder_pipe)
    ])

    # numeric preprocessing
    numeric_pipe = Pipeline([
        ('imputer', SimpleImputer()),
        ('scaler', StandardScaler())
    ])

    # this allows for separate processing of categoric and numeric
    preprocessing = ColumnTransformer(
        [('cat', categoric_pipe, categoric_columns),
         ('num', numeric_pipe, numeric_columns)])

    # final random forest pipeline
    regressor_pipeline = Pipeline([
        ('preprocess', preprocessing),
        ('regressor', regressor)               
    ])
    
    return regressor_pipeline


def TrainXGBRegressor(pipeline, dfX_train, dfy_train, dfX_test=None, dfy_test=None, regressor_args={}):

    """
    Dedicate train function for pipelines with an XGB regressor, supporting early_stopping_rounds

    """


    from sklearn.pipeline import Pipeline
    from xgboost import XGBRegressor

    if 'early_stopping_rounds' in regressor_args.keys() and dfX_test is not None and dfy_test is not None:

        fitted_preprocessor = pipeline['preprocess'].fit(dfX_train, dfy_train)
        dfX_test_trans = fitted_preprocessor.transform(dfX_test)
        fit_params = {'regressor__eval_set':[(dfX_test_trans, dfy_test)],
                      'regressor__verbose':True
                     }
    else:
        fit_params = {}

    model = pipeline.fit(dfX_train, dfy_train, **fit_params)

    return model


def TrainRegressor(pipeline, dfX_train, dfy_train, dfX_test=None, dfy_test=None, regressor_args={}):

    """
    Dedicate train function for regression pipelines

    """

    from sklearn.pipeline import Pipeline

    algo = str(pipeline.named_steps['regressor']).split('(')[0]

    if algo != 'XGBRegressor':

        model = pipeline.fit(dfX_train, dfy_train)

    else:

        if 'early_stopping_rounds' in regressor_args.keys() and dfX_test is not None and dfy_test is not None:

            fitted_preprocessor = pipeline['preprocess'].fit(dfX_train, dfy_train)
            dfX_test_trans = fitted_preprocessor.transform(dfX_test)
            fit_params = {'regressor__eval_set':[(dfX_test_trans, dfy_test)],
                          'regressor__verbose':True
                         }
        else:
            fit_params = {}

        model = pipeline.fit(dfX_train, dfy_train, **fit_params)

    return model


def RegressionGridSearchCV(pipeline,
                           dfX_train, dfy_train,
                           dfX_test=None, dfy_test=None,
                           cv=4,
                           scoring='neg_root_mean_squared_error',
                           regressor_grid_args={}):
    
 
    '''
    Performs a grid search to train the pipeline 
    scoring: 'neg_root_mean_squared_error', 'neg_mean_absolute_error', ...

    input: pipeline, pipeline to be trained using grid search
           dfX_train, dfy_train are dataframes
           dfX_test, dfy_test are only needed for xgb pipelines, dataframes
           scoring = 'neg_root_mean_squared_error' or 'neg_mean_absolute_error'

    output: trained regressor pipeline
    
    '''

    import pandas as pd  
    from sklearn.model_selection import GridSearchCV
    
    
    algo = str(pipeline.named_steps['regressor']).split('(')[0]
    
    if algo == 'RandomForestRegressor':
        
        print('Grid search with RandomForestRegressor')
        
        if not regressor_grid_args:
            pipeline_params = {
                           'preprocess__num__imputer__strategy':['mean','median'],
                           'regressor__n_estimators':[100],
                           'regressor__min_samples_leaf':[1,2,3]
                     }
        else:
            pipeline_params = {
                           'preprocess__num__imputer__strategy':['mean'],
            }
            for k,v in regressor_grid_args.items():
                           pipeline_params['regressor__' + k] = v             
        fit_params = {         
                     }

    elif algo == 'XGBRegressor':
        
        print('Grid search with XGBRegressor')
        
        if not regressor_grid_args:
            pipeline_params = {
                           'regressor__n_estimators':[100],
                           'regressor__min_child_weight':[1,1.5,2],
                           'regressor__learning_rate':[0.3,0.1,0.03],
                           'regressor__max_depth':[5,6],
                           'regressor__use_label_encoder':[False],
                         }
        else:
            pipeline_params = {
                           'preprocess__num__imputer__strategy':['mean'],
            }

            for k,v in regressor_grid_args.items():
                           pipeline_params['regressor__' + k] = v   
            
        fit_params = {         
        #              'classifier__early_stopping_rounds':10,
        #              'classifier__eval_metric':['logloss'],
        #              'classifier__eval_set':[[X_test, y_test]]
                     }
    elif algo == 'Ridge':
        
        print('Grid search with Ridge')
        
        if not regressor_grid_args:
            pipeline_params = {
                           'regressor__alpha':[0.01, 0.1, 1, 10, 100],
                          }
        else:
            pipeline_params = {
                           
            }
            
            for k,v in regressor_grid_args.items():
                           pipeline_params['regressor__' + k] = v   
        fit_params = {
                     }
    elif algo == 'Lasso':
        
        print('Grid search with Lasso')
        
        if not regressor_grid_args:
            pipeline_params = {
                           'regressor__alpha':[0.01, 0.1, 1, 10, 100],
                          }
        else:
            pipeline_params = {
                           
            }
            
            for k,v in regressor_grid_args.items():
                           pipeline_params['regressor__' + k] = v  
        fit_params = {
                     }

    elif algo == 'KerasRegressor':

        print('Grid search with KerasRegressor')
        
        if not regressor_grid_args:
            
            param_grid = {'regressor__model__n_in':[dfX_train.shape[1]],
                          'regressor__model__hidden_layers_sizes':[(1000,),(500,500,),(333,334,333,)],
                          'regressor__optimizer__learning_rate':[0.003,0.01],
                          'regressor__batch_size':[100,1000],
                          'regressor__epochs':[200]
                         }

        else:

            param_grid = {'regressor__model__n_in':[dfX_train.shape[1]]
                         }

            for k,v in regressor_grid_args.items():
                if k in ('hidden_layers_sizes','dropout'):
                    param_grid['regressor__model__' + k] = v
                elif k in ('learning_rate'):
                    param_grid['regressor__optimizer__' + k] = v
                else:
                    param_grid['regressor__' + k] = v
        
    
    if algo != 'KerasRegressor':
        
        grid = GridSearchCV(estimator=pipeline, 
                            param_grid=pipeline_params, 
                            cv=cv,
                            scoring=scoring, 
                            refit=True,
                            verbose=1,
                            n_jobs=1)
    else:

        ####pipeline = CreateClassificationPipeline(dfX_train, max_low_cardinality=0, max_high_cardinality=100000, algo='neuralnet')  
        grid = GridSearchCV(pipeline, param_grid=param_grid, scoring='neg_mean_squared_error')

    import time
    start = time.time()
    trained_regressor = grid.fit(dfX_train, dfy_train) #, **fit_params)
    print("Running time =", time.time() - start, "seconds")

    df_grid_search = _ExtractGridSearchData(grid.cv_results_, regressor_grid_args, 'regression')
        
    return trained_regressor
    


def EvaluateClassifier(model, dfX_test, dfy_test):

    '''
    Computes auc, average precision, accuracy, precision, recall, f1-score
    f1-optimal decision threshold, and accuracy, precision and recall based on
    the f1-optimal decision threshold.
    Returns data for: roc curve, precision-recall curve, f1 v threshold curve, confusion matrix.
    and histogram of predictions.
    
    '''
    
    from sklearn.metrics import confusion_matrix, \
                                precision_score, recall_score, accuracy_score, f1_score, \
                                roc_auc_score, average_precision_score, precision_recall_curve, roc_curve
    
    import pandas as pd
    import numpy as np

    import matplotlib.pyplot as plt
    import seaborn as sns

    y_pred = model.predict_proba(dfX_test)
    auc_score_ = np.round(roc_auc_score(dfy_test, y_pred[:,1]),4)
    average_precision_score_ = np.round(average_precision_score(dfy_test, y_pred[:,1]),4)

    p, r, t = precision_recall_curve(dfy_test, y_pred[:,1])

    p_ratio = dfy_test.mean()
    noskill = p_ratio * np.ones(len(t))

    if PLOTS:
        fig = plt.figure(figsize=(5,5))
        plt.title('Precision-recall curve')
        plt.plot(r,p)
        plt.hlines(p_ratio,0,1,colors='r', linestyles='--', label='no-skill classifier')
        plt.xlim([0,1])
        plt.ylim([0,1])
        plt.ylabel('Precision')
        plt.xlabel('Recall')
        plt.legend()
        plt.show()
    
    df_precision_recall = pd.DataFrame(np.concatenate([t.reshape(-1,1),p[:-1].reshape(-1,1),r[:-1].reshape(-1,1), noskill.reshape(-1,1)], axis=1), 
                                       columns=['threshold','precision','recall','noskill'])

    fpr, tpr, thr = roc_curve(dfy_test, y_pred[:,1])

    if PLOTS:
        fig = plt.figure(figsize=(5,5))
        plt.title('ROC curve')
        plt.plot(fpr, tpr)
        plt.plot([0,1],[0,1], 'r--', label='no-skill classifier')
        plt.xlim([0,1])
        plt.ylim([0,1])
        plt.xlabel('False Positive Rate')
        plt.ylabel('True Positive Rate')
        plt.legend()
        plt.show()
    
    df_roc = pd.DataFrame(np.concatenate([thr.reshape(-1,1), fpr.reshape(-1,1), tpr.reshape(-1,1), fpr.reshape(-1,1)], axis=1), columns=['threshold','fpr','tpr','noskill'])

    y_pred = model.predict(dfX_test)
    precision_score_ = np.round(precision_score(dfy_test, y_pred),4)
    recall_score_ = np.round(recall_score(dfy_test, y_pred),4)
    f1_score_ = np.round(f1_score(dfy_test, y_pred),4)
    accuracy_score_ = np.round(accuracy_score(dfy_test, y_pred),4)

    # compute f1-optimal threshold
    y_pred = model.predict_proba(dfX_test)[:,1]
    df_y_pred = pd.DataFrame(y_pred.reshape(-1,1),columns=['prediction']).astype(float)
    p, r, t = precision_recall_curve(dfy_test, y_pred)
    f1 = np.array([2*x*y/(x+y) for x,y in zip(p,r)])
    f1_optimal_threshold_ = t[np.argmax(f1)]

    if PLOTS:
        fig = plt.figure(figsize=(5,5))
        plt.title('Histogram of y_pred')
        sns.histplot(y_pred, bins=20, edgecolor=None)
        plt.show()

    y_pred[y_pred > f1_optimal_threshold_] = 1
    y_pred[y_pred != 1] = 0

    df_f1_threshold = pd.DataFrame(np.concatenate([t.reshape(-1,1), f1[:-1].reshape(-1,1)], axis=1), columns=['threshold','f1'])
    cm = confusion_matrix(dfy_test.values.reshape(-1,1), y_pred)
    df_confusion_matrix = pd.DataFrame(np.concatenate([np.array(['actual 0','actual 1']).reshape(-1,1), cm], axis=1),
                                       columns=['class', 'predicted 0','predicted 1'])

    if PLOTS:
        fig = plt.figure(figsize=(5,5))
        plt.title('decision-threshold dependence of f1-score')
        plt.plot(t, f1[:-1])
        plt.xlabel('decision-threshold')
        plt.ylabel('f1-score')
        plt.show()
    
    f1_opt_precision_score_ = np.round(precision_score(dfy_test, y_pred),4)
    f1_opt_recall_score_ = np.round(recall_score(dfy_test, y_pred),4)
    f1_opt_f1_score_ = np.round(f1_score(dfy_test, y_pred),4)
    f1_opt_accuracy_score_ = np.round(accuracy_score(dfy_test, y_pred),4)
    
    scores_dict = {}
    scores_dict['auc'] = auc_score_
    scores_dict['average precision'] = average_precision_score_
    scores_dict['precision'] = precision_score_
    scores_dict['recall'] = recall_score_
    scores_dict['accuracy'] = accuracy_score_
    scores_dict['f1'] = f1_score_
    scores_dict['f1-optimal threshold'] = f1_optimal_threshold_
    scores_dict['f1-optimal f1'] = f1_opt_f1_score_
    scores_dict['f1-optimal precision'] = f1_opt_precision_score_
    scores_dict['f1-optimal recall'] = f1_opt_recall_score_
    scores_dict['f1-optimal accuracy'] = f1_opt_accuracy_score_

    df_scores = pd.DataFrame.from_dict(scores_dict, orient='index').reset_index()
    df_scores.columns = ['metric', 'value']
      
    return df_precision_recall, df_roc, df_f1_threshold, df_y_pred, df_confusion_matrix, df_scores


def EvaluateRegressor(model, dfX_test, dfy_test):
    
    '''
    Computes root mean squared error, mean absolute error, explained variance, r-squared,
    and maximum error.

    '''
    
    import numpy as np
    import pandas as pd
    
    from sklearn.metrics import mean_squared_error, mean_absolute_error, \
                                explained_variance_score, r2_score, max_error
    import matplotlib.pyplot as plt
    
    y_pred = model.predict(dfX_test)
    rmse = np.round(mean_squared_error(dfy_test, y_pred, squared=False),4)
    mae = np.round(mean_absolute_error(dfy_test, y_pred),4)
    ev = np.round(explained_variance_score(dfy_test, y_pred),4)
    r2 = np.round(r2_score(dfy_test, y_pred),4)
    me = np.round(max_error(dfy_test, y_pred),4)
    #print('RMSE    = ', rmse)
    
    x_min = min(min(dfy_test),min(y_pred))
    x_max = max(max(dfy_test),max(y_pred))

    # regression evaluation plot and saving the data
    if PLOTS:
        fig = plt.figure(figsize=(8,8))
        plt.scatter(dfy_test, y_pred)
        plt.plot([x_min,x_max],[x_min,x_max], 'r--')
        plt.title('Regression-evaluation plot\nRMSE = ' + str(rmse))
        plt.ylabel('Predicted')
        plt.xlabel('Actual')
        plt.xlim([x_min,x_max])
        plt.ylim([x_min,x_max])
        plt.show()

    data = np.concatenate([dfy_test.values.reshape(-1,1),y_pred.reshape(-1,1)], axis=1)
    df_regression_evaluation = pd.DataFrame(data, columns=['Actual','Predicted'])

    scores_dict = {}
    scores_dict['root mean squared error'] = rmse
    scores_dict['mean absolute error'] = mae
    scores_dict['explained variance'] = ev
    scores_dict['r-squared'] = r2
    scores_dict['max error'] = me

    df_scores = pd.DataFrame.from_dict(scores_dict, orient='index').reset_index()
    df_scores.columns = ['metric', 'value']

    return df_regression_evaluation, df_scores


def ComputeFeatureImportances(estimator, dfX_test, dfy_test, scoring=None):
    
    ''' Computes permutation importances for the estimator using dfX_test and dfy_test
        Computation of the importances is based on the scoring method (default None)
        that is supplied. When the default is used, the estimator's default is used.
        For classifiers that is 'accuracy', for regressors that is 'r2_score'.
    '''
    
    import pandas as pd
    import numpy as np
    from sklearn.inspection import permutation_importance
    
    result = permutation_importance(estimator, dfX_test, dfy_test,
                                    scoring=scoring,
                                    n_repeats=5, random_state=59,
                                    n_jobs=1)
    result.importances_mean[result.importances_mean < 0] = 0

    normalized_importances = np.array(result.importances_mean / np.max(result.importances_mean))
    features = np.array(dfX_test.columns)

    df_importances = pd.DataFrame(np.concatenate([features.reshape(-1,1), normalized_importances.reshape(-1,1)], axis=1),
                                  columns=['feature', 'importance']).sort_values(by='importance', ascending=False)
    df_importances['order'] = np.arange(1, 1+len(features))

    if PLOTS:

        import seaborn as sns
        import matplotlib.pyplot as plt
        
        sns.barplot(x='importance', y='feature',   
                    data=df_importances, 
                    color='red')
        plt.show()

    return df_importances


def ComputeLearningCurve(pipeline, dfX, dfy, mode):

    from sklearn.model_selection import train_test_split
    from sklearn.metrics import roc_auc_score, mean_squared_error
    import pandas as pd

    scores = []
    fractions = (0.1,0.2,0.3,0.4,0.5,0.6,0.7,0.8,0.9)

    for train_size in fractions:

        if mode == 'classification':
            dfX_train, dfX_test, dfy_train, dfy_test = train_test_split(dfX, dfy, stratify=dfy, train_size=train_size, random_state = 59)
        else:
            dfX_train, dfX_test, dfy_train, dfy_test = train_test_split(dfX, dfy, train_size=train_size, random_state = 59)

        model = pipeline.fit(dfX_train, dfy_train)

        if mode == 'classification':
            scores.append(roc_auc_score(dfy_test, model.predict_proba(dfX_test)[:,1]))
        else:
            scores.append(mean_squared_error(dfy_test, model.predict(dfX_test), squared=False))

    df_learning_curve = pd.DataFrame({'fraction':fractions, 'score':scores})

    print(df_learning_curve)

    if PLOTS:

        import matplotlib.pyplot as plt

        fig = plt.figure(figsize=(5,5))
        plt.plot(df_learning_curve['fraction'].values, df_learning_curve['score'])
        plt.title('Learning curve')
        plt.ylabel('Test score')
        plt.xlabel('Fraction of data used for training')
        plt.show()     
            
    return df_learning_curve


def SaveModel(model, modelname=''):

    import cloudpickle

    with open(modelname, 'wb') as fw:
        cloudpickle.dump(model, fw)

    return


def LoadModel(modelname):

    import cloudpickle

    with open(modelname, 'rb') as fr:
        model = cloudpickle.load(fr)

    return model


# ===================================================================================
# EXAMPLE OF HOW TO USE THE FUNCTIONS DEFINED IN THIS SOURCE (simply run the source!)
# ===================================================================================

if __name__ == "__main__":

    import pandas as pd
    from sklearn.model_selection import train_test_split

    PLOTS = False

    import warnings
    warnings.filterwarnings('ignore')

    # =============================
    # BINARY CLASSIFICATION EXAMPLE
    # =============================

    import sklearn
    print('scikit-learn      :', sklearn.__version__)
    print('pandas            :', pd.__version__)
    import xgboost
    print('xgboost           :', xgboost.__version__)
    import numpy as np
    print('numpy             :', np.__version__)
    import category_encoders as ce
    print('category_encoders :', ce.__version__)
    import tensorflow as tf
    print('tensorflow        :', tf.__version__)
    import scikeras
    print('scikeras          :', scikeras.__version__)


    df = pd.read_csv('C:\\Users\\mhorbach\\Datasets\\bankchurners12mar.csv')
    target = 'Attrition_Flag'
      
    dfX = df.drop([target,'CLIENTNUM','Gender'], axis=1)
    dfy = df[target]

    # For now, we require 0/1 labels in the functions, pos_label is set to 1
    dfy_aux = dfy.copy()
    dfy_aux[dfy=='Existing Customer'] = 0
    dfy_aux[dfy=='Attrited Customer'] = 1
    dfy = dfy_aux.astype(int)

    dfX_train, dfX_test, dfy_train, dfy_test = train_test_split(dfX, dfy, stratify=dfy, random_state=59)
   
    MAX_LOW_CARDINALITY = 4
    MAX_HIGH_CARDINALITY = 50


    # define the pipeline . . .
    print('\nXGBOOST:\n')
    classifier_args = {'n_estimators':1000, 'early_stopping_rounds':10,'n_jobs':1}
    pipeline = CreateClassificationPipeline(dfX_train, 
                                            max_low_cardinality=MAX_LOW_CARDINALITY, 
                                            max_high_cardinality=MAX_HIGH_CARDINALITY,
                                            algo='xgboost',
                                            classifier_args=classifier_args)



    # . . . and train it the simple way, and evaluate the result
    print('\nSIMPLE TRAINING WITH EARLY STOPPING...\n')
    model11 = TrainXGBClassifier(pipeline, dfX_train, dfy_train, dfX_test=dfX_test, dfy_test=dfy_test, classifier_args=classifier_args)
    print(model11.named_steps)
    df_precision_recall, df_roc, df_f1_threshold, df_y_pred, df_confusion_matrix, df_scores = EvaluateClassifier(model11, dfX_test, dfy_test)
    df_importances = ComputeFeatureImportances(model11, dfX_test, dfy_test)
    print('\nEVALUATION SCORES:\n')
    print(df_scores)

    # ==================
    # SAVE AND LOAD MODEL
    # ===================

    import cloudpickle
    with open('C:\\Users\\mhorbach\\Desktop\\XGB_saved_model.pkl','wb') as fw:
        cloudpickle.dump(model11,fw)
    with open('C:\\Users\\mhorbach\\Desktop\\XGB_saved_model.pkl','rb') as fr:
        loaded_model = cloudpickle.load(fr)

    y_pred = loaded_model.predict_proba(dfX_train)

    print('First 24 predictions:')
    print(y_pred[:24,1])

    # redo evaluation
    df_precision_recall, df_roc, df_f1_threshold, df_y_pred, df_confusion_matrix, df_scores = EvaluateClassifier(loaded_model, dfX_test, dfy_test)
    df_importances = ComputeFeatureImportances(loaded_model, dfX_test, dfy_test)
    print('\nEVALUATION SCORES:\n')
    print(df_scores)


    # ==================
    # SAVE AND LOAD MODEL
    # ===================

    SaveModel(model11,'C:\\Users\\mhorbach\\Desktop\\XGB_saved_model_SaveModel.pkl')
    loaded_model_2 = LoadModel('C:\\Users\\mhorbach\\Desktop\\XGB_saved_model_SaveModel.pkl')
    
    y_pred = loaded_model_2.predict_proba(dfX_train)

    print('First 24 predictions:')
    print(y_pred[:24,1])

    # redo evaluation
    df_precision_recall, df_roc, df_f1_threshold, df_y_pred, df_confusion_matrix, df_scores = EvaluateClassifier(loaded_model, dfX_test, dfy_test)
    df_importances = ComputeFeatureImportances(loaded_model, dfX_test, dfy_test)
    print('\nEVALUATION SCORES:\n')
    print(df_scores)

    input('WAITING')
    
    
    # ==================
    # REGRESSION EXAMPLE
    # ==================

    df = pd.read_csv('C:\\Users\\mhorbach\\Datasets\\ENB2012.csv')
    target = 'Cooling Load'

    df = df.dropna(subset=[target])
    dfX = df.drop([target, 'Heating Load'], axis=1)
    dfy = df[target]
    

    dfX_train, dfX_test, dfy_train, dfy_test = train_test_split(dfX, dfy, random_state=59)
   
    MAX_LOW_CARDINALITY = 4
    MAX_HIGH_CARDINALITY = 50

    # define the pipeline . . .
    print('\nXGB REGRESSION WITH EARLY STOPPING:\n')
    regressor_args = {'n_estimators':1000, 'early_stopping_rounds':10}
    pipeline = CreateRegressionPipeline(dfX_train, 
                                        max_low_cardinality=MAX_LOW_CARDINALITY, 
                                        max_high_cardinality=MAX_HIGH_CARDINALITY,
                                        algo='xgboost',
                                        regressor_args=regressor_args)

    
    print('\nSIMPLE TRAINING...\n')
    model12 = TrainXGBRegressor(pipeline, dfX_train, dfy_train, dfX_test=dfX_test, dfy_test=dfy_test, regressor_args=regressor_args)
    df_evaluation, scores = EvaluateRegressor(model12, dfX_test, dfy_test)
    df_importances = ComputeFeatureImportances(model12, dfX_test, dfy_test)
    print('\nFEATURE IMPORTANCES:\n')
    print(df_importances)


    # =============================
    # BINARY CLASSIFICATION EXAMPLE
    # =============================

    import sklearn
    print('scikit-learn      :', sklearn.__version__)
    print('pandas            :', pd.__version__)
    import xgboost
    print('xgboost           :', xgboost.__version__)
    import numpy as np
    print('numpy             :', np.__version__)
    import category_encoders as ce
    print('category_encoders :', ce.__version__)
    import tensorflow as tf
    print('tensorflow        :', tf.__version__)
    import scikeras
    print('scikeras          :', scikeras.__version__)


    df = pd.read_csv('C:\\Users\\mhorbach\\Datasets\\bankchurners12mar.csv')
    target = 'Attrition_Flag'
      
    dfX = df.drop(target, axis=1)
    dfy = df[target]

    # For now, we require 0/1 labels in the functions, pos_label is set to 1
    dfy_aux = dfy.copy()
    dfy_aux[dfy=='Existing Customer'] = 0
    dfy_aux[dfy=='Attrited Customer'] = 1
    dfy = dfy_aux.astype(int)

    dfX_train, dfX_test, dfy_train, dfy_test = train_test_split(dfX, dfy, stratify=dfy, random_state=59)
   
    MAX_LOW_CARDINALITY = 4
    MAX_HIGH_CARDINALITY = 50


    # define the pipeline . . .
    print('\nXGBOOST:\n')
    classifier_args = {'n_estimators':1000, 'early_stopping_rounds':10,'n_jobs':1}
    pipeline = CreateClassificationPipeline(dfX_train, 
                                            max_low_cardinality=MAX_LOW_CARDINALITY, 
                                            max_high_cardinality=MAX_HIGH_CARDINALITY,
                                            algo='xgboost',
                                            classifier_args=classifier_args)



    # . . . and train it the simple way, and evaluate the result
    print('\nSIMPLE TRAINING WITH EARLY STOPPING...\n')
    model11 = TrainClassifier(pipeline, dfX_train, dfy_train, dfX_test=dfX_test, dfy_test=dfy_test, classifier_args=classifier_args)
    print(model11.named_steps)
    df_precision_recall, df_roc, df_f1_threshold, df_y_pred, df_confusion_matrix, df_scores = EvaluateClassifier(model11, dfX_test, dfy_test)
    df_importances = ComputeFeatureImportances(model11, dfX_test, dfy_test)
    print('\nEVALUATION SCORES:\n')
    print(df_scores)
    
    # ==================
    # REGRESSION EXAMPLE
    # ==================

    df = pd.read_csv('C:\\Users\\mhorbach\\Datasets\\ENB2012.csv')
    target = 'Cooling Load'

    df = df.dropna(subset=[target])
    dfX = df.drop([target, 'Heating Load'], axis=1)
    dfy = df[target]
    

    dfX_train, dfX_test, dfy_train, dfy_test = train_test_split(dfX, dfy, random_state=59)
   
    MAX_LOW_CARDINALITY = 4
    MAX_HIGH_CARDINALITY = 50

    # define the pipeline . . .
    print('\nXGB REGRESSION WITH EARLY STOPPING:\n')
    regressor_args = {'n_estimators':1000, 'early_stopping_rounds':10}
    pipeline = CreateRegressionPipeline(dfX_train, 
                                        max_low_cardinality=MAX_LOW_CARDINALITY, 
                                        max_high_cardinality=MAX_HIGH_CARDINALITY,
                                        algo='xgboost',
                                        regressor_args=regressor_args)

    
    print('\nSIMPLE TRAINING...\n')
    model12 = TrainRegressor(pipeline, dfX_train, dfy_train, dfX_test=dfX_test, dfy_test=dfy_test, regressor_args=regressor_args)
    df_evaluation, scores = EvaluateRegressor(model12, dfX_test, dfy_test)
    df_importances = ComputeFeatureImportances(model12, dfX_test, dfy_test)
    print('\nFEATURE IMPORTANCES:\n')
    print(df_importances)


    # =============================
    # BINARY CLASSIFICATION EXAMPLE
    # =============================

    import sklearn
    print('scikit-learn      :', sklearn.__version__)
    print('pandas            :', pd.__version__)
    import xgboost
    print('xgboost           :', xgboost.__version__)
    import numpy as np
    print('numpy             :', np.__version__)
    import category_encoders as ce
    print('category_encoders :', ce.__version__)
    import tensorflow as tf
    print('tensorflow        :', tf.__version__)
    import scikeras
    print('scikeras          :', scikeras.__version__)


    df = pd.read_csv('C:\\Users\\mhorbach\\Datasets\\bankchurners12mar.csv')
    target = 'Attrition_Flag'
      
    dfX = df.drop(target, axis=1)
    dfy = df[target]

    # For now, we require 0/1 labels in the functions, pos_label is set to 1
    dfy_aux = dfy.copy()
    dfy_aux[dfy=='Existing Customer'] = 0
    dfy_aux[dfy=='Attrited Customer'] = 1
    dfy = dfy_aux.astype(int)

    dfX_train, dfX_test, dfy_train, dfy_test = train_test_split(dfX, dfy, stratify=dfy, random_state=59)
   
    MAX_LOW_CARDINALITY = 4
    MAX_HIGH_CARDINALITY = 50


    # define the pipeline . . .
    print('\nXGBOOST:\n')
    classifier_args = {'n_jobs':1}
    pipeline = CreateClassificationPipeline(dfX_train, 
                                            max_low_cardinality=MAX_LOW_CARDINALITY, 
                                            max_high_cardinality=MAX_HIGH_CARDINALITY,
                                            algo='randomforest',
                                            classifier_args=classifier_args)



    # . . . and train it the simple way, and evaluate the result
    print('\nSIMPLE TRAINING WITH EARLY STOPPING...\n')
    model11 = TrainClassifier(pipeline, dfX_train, dfy_train, dfX_test=dfX_test, dfy_test=dfy_test, classifier_args=classifier_args)
    print(model11.named_steps)
    df_precision_recall, df_roc, df_f1_threshold, df_y_pred, df_confusion_matrix, df_scores = EvaluateClassifier(model11, dfX_test, dfy_test)
    df_importances = ComputeFeatureImportances(model11, dfX_test, dfy_test)
    print('\nEVALUATION SCORES:\n')
    print(df_scores)
    
    # ==================
    # REGRESSION EXAMPLE
    # ==================

    df = pd.read_csv('C:\\Users\\mhorbach\\Datasets\\ENB2012.csv')
    target = 'Cooling Load'

    df = df.dropna(subset=[target])
    dfX = df.drop([target, 'Heating Load'], axis=1)
    dfy = df[target]
    

    dfX_train, dfX_test, dfy_train, dfy_test = train_test_split(dfX, dfy, random_state=59)
   
    MAX_LOW_CARDINALITY = 4
    MAX_HIGH_CARDINALITY = 50

    # define the pipeline . . .
    print('\nXGB REGRESSION WITH EARLY STOPPING:\n')
    regressor_args = {}
    pipeline = CreateRegressionPipeline(dfX_train, 
                                        max_low_cardinality=MAX_LOW_CARDINALITY, 
                                        max_high_cardinality=MAX_HIGH_CARDINALITY,
                                        algo='randomforest',
                                        regressor_args=regressor_args)

    
    print('\nSIMPLE TRAINING...\n')
    model12 = TrainRegressor(pipeline, dfX_train, dfy_train, dfX_test=dfX_test, dfy_test=dfy_test, regressor_args=regressor_args)
    df_evaluation, scores = EvaluateRegressor(model12, dfX_test, dfy_test)
    df_importances = ComputeFeatureImportances(model12, dfX_test, dfy_test)
    print('\nFEATURE IMPORTANCES:\n')
    print(df_importances)


    input('WAITTTTTTTTTTTTTTTTTTTTTTTTTTTTT')




    # ============================================================
    # BINARY CLASSIFICATION EXAMPLE WITH NEURAL NET + GRIDSEARCHCV
    # ============================================================

    df = pd.read_csv('C:\\Users\\mhorbach\\Datasets\\creditcard.csv')
    target = 'Class'
      
    dfX = df.drop(target, axis=1)
    dfy = df[target]

    # For now, we require 0/1 labels in the functions, pos_label is set to 1
    dfy_aux = dfy.copy()
    #dfy_aux[dfy=='Existing Customer'] = 0
    #dfy_aux[dfy=='Attrited Customer'] = 1
    dfy = dfy_aux.astype(int)

    dfX_train, dfX_test, dfy_train, dfy_test = train_test_split(dfX, dfy, stratify=dfy, random_state=59)
   
    MAX_LOW_CARDINALITY = 4
    MAX_HIGH_CARDINALITY = 50

    # define the pipeline . . . 
    print('\nNEURAL NET:\n')
    pipeline = CreateClassificationPipeline(dfX_train, 
                                            max_low_cardinality=MAX_LOW_CARDINALITY, 
                                            max_high_cardinality=MAX_HIGH_CARDINALITY,
                                            algo='neuralnet',
                                            ) 

    model2 = ClassificationGridSearchCV(pipeline, dfX_train, dfy_train, dfX_test, dfy_test, scoring='roc_auc',
                                        classifier_grid_args={'batch_size':[100],'hidden_layers_sizes':[(50,50,),(25,25,)],
                                                                     'dropout':[0.0],'learning_rate':[0.001]})

    df_precision_recall, df_roc, df_f1_threshold, df_y_pred, df_confusion_matrix, df_scores = EvaluateClassifier(model2, dfX_test, dfy_test)
 
    
  
    # define the pipeline . . .
    print('\nRANDOM FOREST:\n')
    pipeline = CreateClassificationPipeline(dfX_train, 
                                            max_low_cardinality=MAX_LOW_CARDINALITY, 
                                            max_high_cardinality=MAX_HIGH_CARDINALITY,
                                            algo='randomforest',
                                            classifier_args={'n_estimators':5,'min_samples_leaf':2,'n_jobs':1})


    df_learning_curve = ComputeLearningCurve(pipeline, dfX, dfy, 'classification')

    # . . . and train it the simple way, and evaluate the result
    print('\nSIMPLE TRAINING...\n')
    model1 = pipeline.fit(dfX_train, dfy_train)
    print(model1.named_steps)
    df_precision_recall, df_roc, df_f1_threshold, df_y_pred, df_confusion_matrix, df_scores = EvaluateClassifier(model1, dfX_test, dfy_test)
    df_importances = ComputeFeatureImportances(model1, dfX_test, dfy_test)
    print('\nEVALUATION SCORES:\n')
    print(df_scores)
    print('\nTOP OF PRECISION-RECALL CURVE DATA:\n')
    print(df_precision_recall.head(12))
    print('\nTOP OF ROC CURVE DATA:\n')
    print(df_roc.head(12))
    print('\nFEATURE IMPORTANCES:\n')
    print(df_importances)
    print(model1.named_steps)
    
    # . . . or train it using a grid search, and evaluate the result
    print('\nGRID SEARCH...\n')
    model2 = ClassificationGridSearchCV(pipeline, dfX_train, dfy_train, dfX_test, dfy_test, scoring='roc_auc')
    df_precision_recall, df_roc, df_f1_threshold, df_y_pred, df_confusion_matrix, df_scores = EvaluateClassifier(model2, dfX_test, dfy_test)
    df_importances = ComputeFeatureImportances(model2, dfX_test, dfy_test)
    print('\nEVALUATION SCORES:\n')
    print(df_scores)
    print('\nTOP OF PRECISION-RECALL CURVE DATA:\n')
    print(df_precision_recall.head(12))
    print('\nTOP OF ROC CURVE DATA:\n')
    print(df_roc.head(12))
    print('\nFEATURE IMPORTANCES:\n')
    print(df_importances)

    # . . . or train it using a grid search, and evaluate the result
    print('\nGRID SEARCH WITH USER-SPECIFIED GRID...\n')
    model3 = ClassificationGridSearchCV(pipeline, dfX_train, dfy_train, dfX_test, dfy_test, scoring='roc_auc',
                                        classifier_grid_args={'n_estimators':[20,40,60,80,100],'min_samples_leaf':[1,2,3,4,5],'n_jobs':[1]})
    df_precision_recall, df_roc, df_f1_threshold, df_y_pred, df_confusion_matrix, df_scores = EvaluateClassifier(model3, dfX_test, dfy_test)
    df_importances = ComputeFeatureImportances(model3, dfX_test, dfy_test)
    print('\nEVALUATION SCORES:\n')
    print(df_scores)
    print('\nTOP OF PRECISION-RECALL CURVE DATA:\n')
    print(df_precision_recall.head(12))
    print('\nTOP OF ROC CURVE DATA:\n')
    print(df_roc.head(12))
    print('\nFEATURE IMPORTANCES:\n')
    print(df_importances)

    print('\nXGBOOST:\n')
    # define the pipeline . . .
    pipeline = CreateClassificationPipeline(dfX_train, 
                                            max_low_cardinality=MAX_LOW_CARDINALITY, 
                                            max_high_cardinality=MAX_HIGH_CARDINALITY,
                                            algo='xgboost',
                                            classifier_args={'n_estimators':5,'eval_metric':'nonsense','n_jobs':1})

    # . . . and train it the simple way, and evaluate the result
    print('\nSIMPLE TRAINING...\n')
    model1 = pipeline.fit(dfX_train, dfy_train)
    print(model1.named_steps)
    df_precision_recall, df_roc, df_f1_threshold, df_y_pred, df_confusion_matrix, df_scores = EvaluateClassifier(model1, dfX_test, dfy_test)
    df_importances = ComputeFeatureImportances(model1, dfX_test, dfy_test)
    print('\nEVALUATION SCORES:\n')
    print(df_scores)
    print('\nTOP OF PRECISION-RECALL CURVE DATA:\n')
    print(df_precision_recall.head(12))
    print('\nTOP OF ROC CURVE DATA:\n')
    print(df_roc.head(12))
    print('\nFEATURE IMPORTANCES:\n')
    print(df_importances)
    print(model1.named_steps)

    # . . . or train it using a grid search, and evaluate the result
    print('\nGRID SEARCH...\n')
    model2 = ClassificationGridSearchCV(pipeline, dfX_train, dfy_train, dfX_test, dfy_test, scoring='roc_auc')
    df_precision_recall, df_roc, df_f1_threshold, df_y_pred, df_confusion_matrix, df_scores = EvaluateClassifier(model2, dfX_test, dfy_test)
    df_importances = ComputeFeatureImportances(model2, dfX_test, dfy_test)
    print('\nEVALUATION SCORES:\n')
    print(df_scores)
    print('\nTOP OF PRECISION-RECALL CURVE DATA:\n')
    print(df_precision_recall.head(12))
    print('\nTOP OF ROC CURVE DATA:\n')
    print(df_roc.head(12))
    print('\nFEATURE IMPORTANCES:\n')
    print(df_importances)

    # . . . or train it using a grid search, and evaluate the result
    print('\nGRID SEARCH WITH USER-SPECIFIED GRID...\n')
    model3 = ClassificationGridSearchCV(pipeline, dfX_train, dfy_train, dfX_test, dfy_test, scoring='roc_auc',
                                        classifier_grid_args={'n_estimators':[2,100,400]})
    df_precision_recall, df_roc, df_f1_threshold, df_y_pred, df_confusion_matrix, df_scores = EvaluateClassifier(model3, dfX_test, dfy_test)
    df_importances = ComputeFeatureImportances(model3, dfX_test, dfy_test)
    print('\nEVALUATION SCORES:\n')
    print(df_scores)
    print('\nTOP OF PRECISION-RECALL CURVE DATA:\n')
    print(df_precision_recall.head(12))
    print('\nTOP OF ROC CURVE DATA:\n')
    print(df_roc.head(12))
    print('\nFEATURE IMPORTANCES:\n')
    print(df_importances)

    print('\nLOGISTIC REGRESSION:\n')
    # define the pipeline . . .
    pipeline = CreateClassificationPipeline(dfX_train, 
                                            max_low_cardinality=MAX_LOW_CARDINALITY, 
                                            max_high_cardinality=MAX_HIGH_CARDINALITY,
                                            algo='logisticregression',
                                            classifier_args={'C':1,'n_jobs':1})

    # . . . and train it the simple way, and evaluate the result
    print('\nSIMPLE TRAINING...\n')
    model1 = pipeline.fit(dfX_train, dfy_train)
    print(model1.named_steps)
    df_precision_recall, df_roc, df_f1_threshold, df_y_pred, df_confusion_matrix, df_scores = EvaluateClassifier(model1, dfX_test, dfy_test)
    df_importances = ComputeFeatureImportances(model1, dfX_test, dfy_test)
    print('\nEVALUATION SCORES:\n')
    print(df_scores)
    print('\nTOP OF PRECISION-RECALL CURVE DATA:\n')
    print(df_precision_recall.head(12))
    print('\nTOP OF ROC CURVE DATA:\n')
    print(df_roc.head(12))
    print('\nFEATURE IMPORTANCES:\n')
    print(df_importances)
    print(model1.named_steps)

    # . . . or train it using a grid search, and evaluate the result
    print('\nGRID SEARCH...\n')
    model2 = ClassificationGridSearchCV(pipeline, dfX_train, dfy_train, dfX_test, dfy_test, scoring='roc_auc')
    df_precision_recall, df_roc, df_f1_threshold, df_y_pred, df_confusion_matrix, df_scores = EvaluateClassifier(model2, dfX_test, dfy_test)
    df_importances = ComputeFeatureImportances(model2, dfX_test, dfy_test)
    print('\nEVALUATION SCORES:\n')
    print(df_scores)
    print('\nTOP OF PRECISION-RECALL CURVE DATA:\n')
    print(df_precision_recall.head(12))
    print('\nTOP OF ROC CURVE DATA:\n')
    print(df_roc.head(12))
    print('\nFEATURE IMPORTANCES:\n')
    print(df_importances)

   # . . . or train it using a grid search, and evaluate the result
    print('\nGRID SEARCH WITH USER-SPECIFIED GRID...\n')
    model3 = ClassificationGridSearchCV(pipeline, dfX_train, dfy_train, dfX_test, dfy_test, scoring='roc_auc',
                                        classifier_grid_args={'C':[0.1,1,0000000.1],'n_jobs':[1]})
    df_precision_recall, df_roc, df_f1_threshold, df_y_pred, df_confusion_matrix, df_scores = EvaluateClassifier(model3, dfX_test, dfy_test)
    df_importances = ComputeFeatureImportances(model3, dfX_test, dfy_test)
    print('\nEVALUATION SCORES:\n')
    print(df_scores)
    print('\nTOP OF PRECISION-RECALL CURVE DATA:\n')
    print(df_precision_recall.head(12))
    print('\nTOP OF ROC CURVE DATA:\n')
    print(df_roc.head(12))
    print('\nFEATURE IMPORTANCES:\n')
    print(df_importances)
    

    # ==================
    # REGRESSION EXAMPLE
    # ==================

    df = pd.read_csv('C:\\Users\\mhorbach\\Datasets\\ENB2012.csv')
    target = 'Cooling Load'

    df = df.dropna(subset=[target])
    dfX = df.drop([target, 'Heating Load'], axis=1)
    dfy = df[target]
    

    dfX_train, dfX_test, dfy_train, dfy_test = train_test_split(dfX, dfy, random_state=59)
   
    MAX_LOW_CARDINALITY = 4
    MAX_HIGH_CARDINALITY = 50

    # define the pipeline . . .
    print('\nRANDOM FOREST:\n')
    pipeline = CreateRegressionPipeline(dfX_train, 
                                        max_low_cardinality=MAX_LOW_CARDINALITY, 
                                        max_high_cardinality=MAX_HIGH_CARDINALITY,
                                        algo='randomforest',
                                        regressor_args={'n_estimators':10,'n_jobs':1})

    ComputeLearningCurve(pipeline, dfX, dfy, 'regression')

    # . . . and train it using a grid search, and evaluate the result
    print('\nGRID SEARCH...\n')
    model1 = RegressionGridSearchCV(pipeline, dfX_train, dfy_train, dfX_test, dfy_test,
                                    scoring='neg_mean_squared_error')
    df_evaluation, scores = EvaluateRegressor(model1, dfX_test, dfy_test)
    df_importances = ComputeFeatureImportances(model1, dfX_test, dfy_test)
    print('\nEVALUATION SCORES:\n')
    print(scores)
    print('\nTOP OF EVALUATION PLOT DATA:\n')
    print(df_evaluation.head(12))
    print('\nFEATURE IMPORTANCES:\n')
    print(df_importances)

    # define the pipeline . . .
    print('\nXGBOOST:\n')
    pipeline = CreateRegressionPipeline(dfX_train, 
                                        max_low_cardinality=MAX_LOW_CARDINALITY, 
                                        max_high_cardinality=MAX_HIGH_CARDINALITY,
                                        algo='xgboost',
                                        regressor_args={'n_estimators':10})

    # . . . and train it using a grid search, and evaluate the result
    print('\nGRID SEARCH...\n')
    model1 = RegressionGridSearchCV(pipeline, dfX_train, dfy_train, dfX_test, dfy_test,
                                    scoring='neg_mean_squared_error')
    df_evaluation, scores = EvaluateRegressor(model1, dfX_test, dfy_test)
    df_importances = ComputeFeatureImportances(model1, dfX_test, dfy_test)
    print('\nEVALUATION SCORES:\n')
    print(scores)
    print('\nTOP OF EVALUATION PLOT DATA:\n')
    print(df_evaluation.head(12))
    print('\nFEATURE IMPORTANCES:\n')
    print(df_importances)

    # define the pipeline . . .
    print('\nRIDGE:\n')
    pipeline = CreateRegressionPipeline(dfX_train, 
                                        max_low_cardinality=MAX_LOW_CARDINALITY, 
                                        max_high_cardinality=MAX_HIGH_CARDINALITY,
                                        algo='ridge',
                                        regressor_args={'alpha':10})

    # . . . and train it using a grid search, and evaluate the result
    print('\nGRID SEARCH...\n')
    model1 = RegressionGridSearchCV(pipeline, dfX_train, dfy_train, dfX_test, dfy_test,
                                    scoring='neg_mean_squared_error')
    df_evaluation, scores = EvaluateRegressor(model1, dfX_test, dfy_test)
    df_importances = ComputeFeatureImportances(model1, dfX_test, dfy_test)
    print('\nEVALUATION SCORES:\n')
    print(scores)
    print('\nTOP OF EVALUATION PLOT DATA:\n')
    print(df_evaluation.head(12))
    print('\nFEATURE IMPORTANCES:\n')
    print(df_importances)

    # define the pipeline . . .
    print('\nLASSO:\n')
    pipeline = CreateRegressionPipeline(dfX_train, 
                                        max_low_cardinality=MAX_LOW_CARDINALITY, 
                                        max_high_cardinality=MAX_HIGH_CARDINALITY,
                                        algo='lasso',
                                        regressor_args={'alpha':10})

    # . . . and train it using a grid search, and evaluate the result
    print('\nGRID SEARCH...\n')
    model1 = RegressionGridSearchCV(pipeline, dfX_train, dfy_train, dfX_test, dfy_test,
                                    scoring='neg_mean_squared_error')
    df_evaluation, scores = EvaluateRegressor(model1, dfX_test, dfy_test)
    df_importances = ComputeFeatureImportances(model1, dfX_test, dfy_test)
    print('\nEVALUATION SCORES:\n')
    print(scores)
    print('\nTOP OF EVALUATION PLOT DATA:\n')
    print(df_evaluation.head(12))
    print('\nFEATURE IMPORTANCES:\n')
    print(df_importances)


    df = pd.read_csv('C:\\Users\\mhorbach\\Datasets\\bankchurners12mar.csv')
    target = 'Attrition_Flag'
      
    dfX = df.drop(target, axis=1)
    dfy = df[target]

    # For now, we require 0/1 labels in the functions, pos_label is set to 1
    dfy_aux = dfy.copy()
    dfy_aux[dfy=='Existing Customer'] = 0
    dfy_aux[dfy=='Attrited Customer'] = 1
    dfy = dfy_aux.astype(int)

    dfX_train, dfX_test, dfy_train, dfy_test = train_test_split(dfX, dfy, stratify=dfy, random_state=59)


    # define the pipeline . . .
    print('\nCARDINALITY BOUNDARY CASES - NO ENCODING AT ALL:\n')
    pipeline = CreateClassificationPipeline(dfX_train, 
                                            max_low_cardinality=1, 
                                            max_high_cardinality=1,
                                            algo='randomforest',
                                            classifier_args={'n_estimators':5,'min_samples_leaf':2,'n_jobs':1})

    # . . . and train it the simple way, and evaluate the result
    print('\nSIMPLE TRAINING...\n')
    model1 = pipeline.fit(dfX_train, dfy_train)
    print(model1.named_steps)
    df_precision_recall, df_roc, df_f1_threshold, df_y_pred, df_confusion_matrix, df_scores = EvaluateClassifier(model1, dfX_test, dfy_test)
    df_importances = ComputeFeatureImportances(model1, dfX_test, dfy_test)
    print('\nEVALUATION SCORES:\n')
    print(df_scores)
    print('\nTOP OF PRECISION-RECALL CURVE DATA:\n')
    print(df_precision_recall.head(12))
    print('\nTOP OF ROC CURVE DATA:\n')
    print(df_roc.head(12))
    print('\nFEATURE IMPORTANCES:\n')
    print(df_importances)
    print(model1.named_steps)
    
    # . . . or train it using a grid search, and evaluate the result
    print('\nGRID SEARCH...\n')
    model2 = ClassificationGridSearchCV(pipeline, dfX_train, dfy_train, dfX_test, dfy_test, scoring='roc_auc')
    df_precision_recall, df_roc, df_f1_threshold, df_y_pred, df_confusion_matrix, df_scores = EvaluateClassifier(model2, dfX_test, dfy_test)
    df_importances = ComputeFeatureImportances(model2, dfX_test, dfy_test)
    print('\nEVALUATION SCORES:\n')
    print(df_scores)
    print('\nTOP OF PRECISION-RECALL CURVE DATA:\n')
    print(df_precision_recall.head(12))
    print('\nTOP OF ROC CURVE DATA:\n')
    print(df_roc.head(12))
    print('\nFEATURE IMPORTANCES:\n')
    print(df_importances)


    # ==================================================
    # REGRESSION EXAMPLE TO TEST HIGH CARDINALITY SCALER
    # ==================================================

    df = pd.read_csv('C:\\Users\\mhorbach\\Datasets\\bankchurners.csv')
    target = 'Total_Trans_Ct'

    df = df.dropna(subset=[target])
    dfX = df.drop([target], axis=1)
    dfy = df[target]
    

    dfX_train, dfX_test, dfy_train, dfy_test = train_test_split(dfX, dfy, random_state=59)
   
    MAX_LOW_CARDINALITY = 4
    MAX_HIGH_CARDINALITY = 50

    # define the pipeline . . .
    print('\nRANDOM FOREST:\n')
    pipeline = CreateRegressionPipeline(dfX_train, 
                                        max_low_cardinality=MAX_LOW_CARDINALITY, 
                                        max_high_cardinality=MAX_HIGH_CARDINALITY,
                                        algo='randomforest',
                                        regressor_args={'n_estimators':10,'n_jobs':1})

    # . . . and train it using a grid search, and evaluate the result
    print('\nGRID SEARCH...\n')
    model1 = RegressionGridSearchCV(pipeline, dfX_train, dfy_train, dfX_test, dfy_test,
                                    scoring='neg_mean_squared_error')
    df_evaluation, scores = EvaluateRegressor(model1, dfX_test, dfy_test)
    df_importances = ComputeFeatureImportances(model1, dfX_test, dfy_test)
    print('\nEVALUATION SCORES:\n')
    print(scores)
    print('\nTOP OF EVALUATION PLOT DATA:\n')
    print(df_evaluation.head(12))
    print('\nFEATURE IMPORTANCES:\n')
    print(df_importances)


    # ==================================
    # REGRESSION EXAMPLE WITH NEURAL NET
    # ==================================

    #df = pd.read_csv('C:\\Users\\mhorbach\\Datasets\\ENB2012.csv')
    #target = 'Cooling Load'

    #df = df.dropna(subset=[target])
    #dfX = df.drop([target, 'Heating Load'], axis=1)
    #dfy = df[target]

    df = pd.read_csv('C:\\Users\\mhorbach\\Datasets\\kaggle_cars.csv')
    target = 'price'

    #df = df.dropna(subset=[target])
    dfX = df.drop([target, 'tax'], axis=1)
    dfy = df[target]
    

    dfX_train, dfX_test, dfy_train, dfy_test = train_test_split(dfX, dfy, random_state=59)
   
    MAX_LOW_CARDINALITY = 4
    MAX_HIGH_CARDINALITY = 50

    # define the pipeline . . .
    print('\nNEURALNET WITH DEFAULT PARMS:\n')
    pipeline = CreateRegressionPipeline(dfX_train, 
                                        max_low_cardinality=MAX_LOW_CARDINALITY, 
                                        max_high_cardinality=MAX_HIGH_CARDINALITY,
                                        algo='neuralnet', regressor_args={'epochs':100})

    df_learning_curve = ComputeLearningCurve(pipeline, dfX, dfy, 'regression')

    # straight fit
    model0 = pipeline.fit(dfX_train, dfy_train)
    df_evaluation, scores = EvaluateRegressor(model0, dfX_test, dfy_test)

    # define the pipeline . . .
    print('\nNEURALNET WITH USER-PARMS:\n')
    pipeline = CreateRegressionPipeline(dfX_train, 
                                        max_low_cardinality=MAX_LOW_CARDINALITY, 
                                        max_high_cardinality=MAX_HIGH_CARDINALITY,
                                        algo='neuralnet',
                                        regressor_args={'hidden_layers_sizes':(50,50,),'learning_rate':0.002,'epochs':100})


    # straight fit
    model0 = pipeline.fit(dfX_train, dfy_train)
    df_evaluation, scores = EvaluateRegressor(model0, dfX_test, dfy_test)


    # ==================================
    # REGRESSION EXAMPLE WITH NEURAL NET DEFAULT GRID
    # ==================================

    df = pd.read_csv('C:\\Users\\mhorbach\\Datasets\\ENB2012.csv')
    target = 'Cooling Load'

    df = df.dropna(subset=[target])
    dfX = df.drop([target, 'Heating Load'], axis=1)
    dfy = df[target]

    dfX_train, dfX_test, dfy_train, dfy_test = train_test_split(dfX, dfy, random_state=59)
   
    MAX_LOW_CARDINALITY = 4
    MAX_HIGH_CARDINALITY = 50

    # define the pipeline . . .
    print('\nNEURALNET:\n')
    pipeline = CreateRegressionPipeline(dfX_train, 
                                        max_low_cardinality=MAX_LOW_CARDINALITY, 
                                        max_high_cardinality=MAX_HIGH_CARDINALITY,
                                        algo='neuralnet')#, #,
                                        #regressor_args={'n_hidden':3,'n_neurons':200})

    # . . . and train it using a grid search, and evaluate the result
    print('\nGRID SEARCH...\n')
    model1 = RegressionGridSearchCV(pipeline,
                                    dfX_train, dfy_train,
                                    dfX_test, dfy_test,
                                    scoring='neg_mean_squared_error',
                                    #regressor_grid_args={'model__n_hidden':[3,4],'model__n_neurons':[100,400]}
                                    )
    
    df_evaluation, scores = EvaluateRegressor(model1, dfX_test, dfy_test)


    # ==================================
    # REGRESSION EXAMPLE WITH NEURAL NET USER GRID
    # ==================================

    df = pd.read_csv('C:\\Users\\mhorbach\\Datasets\\ENB2012.csv')
    target = 'Cooling Load'

    df = df.dropna(subset=[target])
    dfX = df.drop([target, 'Heating Load'], axis=1)
    dfy = df[target]

    dfX_train, dfX_test, dfy_train, dfy_test = train_test_split(dfX, dfy, random_state=59)
   
    MAX_LOW_CARDINALITY = 4
    MAX_HIGH_CARDINALITY = 50

    # define the pipeline . . .
    print('\nNEURALNET:\n')
    pipeline = CreateRegressionPipeline(dfX_train, 
                                        max_low_cardinality=MAX_LOW_CARDINALITY, 
                                        max_high_cardinality=MAX_HIGH_CARDINALITY,
                                        algo='neuralnet') #,
                                        #regressor_args={'n_hidden':3,'n_neurons':200})

    # . . . and train it using a grid search, and evaluate the result
    print('\nGRID SEARCH...\n')
    model1 = RegressionGridSearchCV(pipeline,
                                    dfX_train, dfy_train,
                                    dfX_test, dfy_test,
                                    scoring='neg_mean_squared_error',
                                    regressor_grid_args={'model__hidden_layers_sizes':[(1,),(50,50,),(100,100,)], 'optimizer__learning_rate':[0.0012, 0.00000001], 'epochs':[201], 'batch_size':[64]},
                                    )
    
    df_evaluation, scores = EvaluateRegressor(model1, dfX_test, dfy_test)

    
    # =============================================
    # BINARY CLASSIFICATION EXAMPLE WITH NEURAL NET
    # =============================================

    df = pd.read_csv('C:\\Users\\mhorbach\\Datasets\\bankchurners12mar.csv')
    target = 'Attrition_Flag'
      
    dfX = df.drop(target, axis=1)
    dfy = df[target]

    # For now, we require 0/1 labels in the functions, pos_label is set to 1
    dfy_aux = dfy.copy()
    dfy_aux[dfy=='Existing Customer'] = 0
    dfy_aux[dfy=='Attrited Customer'] = 1
    dfy = dfy_aux.astype(int)

    dfX_train, dfX_test, dfy_train, dfy_test = train_test_split(dfX, dfy, stratify=dfy, random_state=59)
   
    MAX_LOW_CARDINALITY = 4
    MAX_HIGH_CARDINALITY = 50

    # define the pipeline . . . 
    print('\nNEURAL NET:\n')
    pipeline = CreateClassificationPipeline(dfX_train, 
                                            max_low_cardinality=MAX_LOW_CARDINALITY, 
                                            max_high_cardinality=MAX_HIGH_CARDINALITY,
                                            algo='neuralnet',
                                            classifier_args={'patience':7,'epochs':88,'batch_size':22,'hidden_layers_sizes':(25,25,25,),'dropout':0,'learning_rate':0.002}
                                            ) # . . . and train it the simple way, and evaluate the result
    print('\nSIMPLE TRAINING...\n')
    model1 = pipeline.fit(dfX_train, dfy_train)
    print(model1.named_steps)
    df_precision_recall, df_roc, df_f1_threshold, df_y_pred, df_confusion_matrix, df_scores = EvaluateClassifier(model1, dfX_test, dfy_test)

    if False:
        df_importances = ComputeFeatureImportances(model1, dfX_test, dfy_test)
        print('\nFEATURE IMPORTANCES:\n')
        print(df_importances)

    print('\nEVALUATION SCORES:\n')   
    print(df_scores)
    print('\nTOP OF PRECISION-RECALL CURVE DATA:\n')
    print(df_precision_recall.head(12))
    print('\nTOP OF ROC CURVE DATA:\n')
    print(df_roc.head(12))

    

     # . . . and train it using a grid search, and evaluate the result
    print('\nGRID SEARCH WITH DEFAULT GRID ...\n')
    model1 = ClassificationGridSearchCV(pipeline,
                                    dfX_train, dfy_train,
                                    dfX_test, dfy_test,
                                    scoring='roc_auc',
                                    #classifier_grid_args={'model__n_hidden':[2],'model__n_neurons':[40,401], 'optimizer__learning_rate':[0.0015]},
                                    cv=10
                                    )

    print('CHECK THE GRID PARAMS NOW:', model1.best_estimator_) ##############???
    print('RESULTS:', model1.cv_results_)
    df_precision_recall, df_roc, df_f1_threshold, df_y_pred, df_confusion_matrix, df_scores = EvaluateClassifier(model1, dfX_test, dfy_test)
    print('\nEVALUATION SCORES:\n')   
    print(df_scores)


    print('\nGRID SEARCH WITH USER-SPECIFIED GRID ...\n')
    model1 = ClassificationGridSearchCV(pipeline,
                                    dfX_train, dfy_train,
                                    dfX_test, dfy_test,
                                    scoring='roc_auc',
                                    classifier_grid_args={'model__hidden_layers_sizes':[(1,),(50,50,),(100,100,)], 'optimizer__learning_rate':[0.0012, 0.00000001]},
                                    cv=10
                                    )

    print('CHECK THE GRID PARAMS NOW:', model1.best_estimator_) 
    print('RESULTS:', model1.cv_results_)
    df_precision_recall, df_roc, df_f1_threshold, df_y_pred, df_confusion_matrix, df_scores = EvaluateClassifier(model1, dfX_test, dfy_test)
    print('\nEVALUATION SCORES:\n')   
    print(df_scores)

    # =============================================
    # BINARY CLASSIFICATION EXAMPLE WITH NEURAL NET
    # =============================================

    df = pd.read_csv('C:\\Users\\mhorbach\\Datasets\\creditcard.csv')
    target = 'Class'
      
    dfX = df.drop(target, axis=1)
    dfy = df[target]

    # For now, we require 0/1 labels in the functions, pos_label is set to 1
    dfy_aux = dfy.copy()
    #dfy_aux[dfy=='Existing Customer'] = 0
    #dfy_aux[dfy=='Attrited Customer'] = 1
    dfy = dfy_aux.astype(int)

    dfX_train, dfX_test, dfy_train, dfy_test = train_test_split(dfX, dfy, stratify=dfy, random_state=59)
   
    MAX_LOW_CARDINALITY = 4
    MAX_HIGH_CARDINALITY = 50

    # define the pipeline . . . 
    print('\nNEURAL NET:\n')
    pipeline = CreateClassificationPipeline(dfX_train, 
                                            max_low_cardinality=MAX_LOW_CARDINALITY, 
                                            max_high_cardinality=MAX_HIGH_CARDINALITY,
                                            algo='neuralnet',
                                            classifier_args={'patience':5,'epochs':100,'batch_size':1000,'hidden_layers_sizes':(50,50,50,),'dropout':0.0,'learning_rate':0.001}
                                            ) # . . . and train it the simple way, and evaluate the result
    print('\nSIMPLE TRAINING...\n')
    model1 = pipeline.fit(dfX_train, dfy_train)
    print(model1.named_steps)
    df_precision_recall, df_roc, df_f1_threshold, df_y_pred, df_confusion_matrix, df_scores = EvaluateClassifier(model1, dfX_test, dfy_test)
 
    PLOTS = False


 
    
    






    
    

  

    

    



#
# --- LIBRARIES ---------------------------------------------------------------
import numpy as np
import pandas as pd
import json
#work in progress, not currently used
#import base64
#import collections
#import random
#import re
#import requests
#import warnings

#spotfire-dsml modules
from spotfire_dsml.ml_metadata import metadata_data as mdd


# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
# Scorer classes

class BaseModelScorer():
    '''
    A base implementation for a model Scorer.
    '''

    # class variables ----------------------------------------------------------
    # Name of column containing prediction
    prediction_col = 'prediction'
    #: str: prefix for json schema parms description
    worker_parms_description_prefix = 'init parameters for class '

    # class methods -----------------------------------------------------------
    @classmethod
    def choose_scorer(cls, model_origin, model_type, target=None, target_orig_classes=[], target_int_values=[]):
        '''
        Class method.
        Depending on the value of model_origin, initialise the correct
        Scorer subclass. If the model_origin or model_type are not implemented,
        raise a NotImplemented error.

        Parameters
        ----------
        model_origin : str
            The source of the model. Currently only "sklearn" implemented.
        model_type : str
            One of "classification" or "regression".
        target : str, optional
            The name of the target variable. The default is None.
        target_orig_classes : list of str, optional
            The list of class names, if model_type is "classification".
            The default is [].
        target_int_values : list of int, optional
            The list of class integer values, if model_type is "classification".
            The default is [].

        Raises
        ------
        NotImplementedError
            If the model_origin is not implemented, raise
            a NotImplemented error stating that the model_origin is
            not implemented yet.
            Likewise, if the model_type is not one of "classification"
            or "regression".

        Returns
        -------
        Subclass of BaseModelScorer
            Currently returns an instance of SklearnScorer, or None if
            model_origin not "sklearn".

        '''
        if not model_type in ['classification','regression']:
            raise NotImplementedError(str(cls)+'.choose_scorer: The '+model_type+'model_type is not implemented yet')
            return None

        if model_origin == 'sklearn':
            return SklearnScorer(model_type, target, target_orig_classes, target_int_values)
        else:
            raise NotImplementedError(str(cls)+'.choose_scorer: The '+model_origin+'model_origin is not implemented yet')
            return None

    @classmethod
    def unpack_json(cls, received_json):
        '''
        Class method. Unpack a JSON object into its original Scorer object.

        Parameters
        ----------
        received_json : dictionary or string
            The JSON object to unpack. If the object is in string format,
            it will be automatically loaded into a dictionary.

        Raises
        ------
        KeyError
            If the schema, the parameters or the metadata for this Worker were not
            found in the JSON object.

        Returns
        -------
        Scorer object.
            The unpacked Scorer object.

        '''

        #If the json is not in dictionary form, turn it into one
        if type(received_json)== str:
            received_json = json.loads(received_json)

        try:
            schema = json.loads(received_json['schema'])
        except KeyError:
            raise KeyError(
                str(cls)+'.unpack_json: The schema key was not in json dictionary')

        # Find class name (remove whatever comes before last dot)
        my_class_name = cls.__name__.split('.')[-1]
        # Construct description strings (same as in function pack_json)
        key_parms_search = cls.worker_parms_description_prefix + my_class_name
        # Find them in the schema dictionary
        parms_key = list(filter(lambda x: key_parms_search in x, schema))

        if len(parms_key) > 0:
            parms = schema[parms_key[0]]
            scorer_attributes = json.loads(received_json[parms])
            # Create a new score object
            scorer = cls.choose_scorer(**scorer_attributes)
        else:
            raise KeyError(
                str(cls)+'.unpack_json: The parms parameter was not in json dictionary: '+key_parms_search)

        return scorer

    # object methods -----------------------------------------------------------

    def pack_json(self, scorer_parms_name):
        '''
        Take all information from this object, and turn it into
        JSON.

        Parameters
        ----------
        scorer_parms_name : str
            The name for the scorer parameters. This will be stored in
            json_schema to be retrieved in json_body.

        Returns
        -------
        json_schema : dictionary
            The json minischema.
        json_body : dictionary
            The json body.

        '''
        json_parms = self.init_attributes_to_json()
        # Find class name (remove whatever comes before last dot)
        # ..but find the base class because the unpack method is class-based on the parent class
        immediate_parent = self.__class__.__bases__[0]
        my_class_name = immediate_parent.__name__.split('.')[-1]
        # Create labels and mini schema
        worker_parms_description = self.__class__.worker_parms_description_prefix + my_class_name
        json_schema = {worker_parms_description: scorer_parms_name}
        # Create body of json
        json_body = {scorer_parms_name: json_parms}
        return (json_schema, json_body)


    def __init__(self, model_origin, model_type, target=None, target_orig_classes=[], target_int_values=[]):
        '''
        Initialize object by storing the relevant information.
        Save instance attributes passed in into special
        instance variable __init_attributes__.

        Parameters
        ----------
        model_origin : str
            The source of the model. Currently only "sklearn" implemented.
        model_type : str
            One of "classification" or "regression".
        target : str, optional
            The name of the target variable. The default is None.
        target_orig_classes : list of str, optional
            The list of class names, if model_type is "classification".
            The default is [].
        target_int_values : list of int, optional
            The list of class integer values, if model_type is "classification".
            The default is [].

        Returns
        -------
        None.

        '''
        self.model_origin = model_origin
        self.model_type = model_type
        self.target = target
        self.target_orig_classes=target_orig_classes
        self.target_int_values=[int(x) for x in target_int_values]

        # save a copy of the instance attributes that were defined during construction
        self.__init_attributes__ = self.__dict__.copy()

        return None

    def init_attributes_to_json(self):
        '''
        Collect the instance attributes defined at construction,
        and turn them into a json string.
        Used within a constructor method.

        Returns
        -------
        string
            JSON string containing instance attributes needed to
            configure scorer.

        '''
        return json.dumps(self.__init_attributes__)

    def setup(self, model):
        '''
        Store predictive model as instance variable

        Parameters
        ----------
        model : bytes or sklearn pipeline
            The predictive model object.

        Returns
        -------
        self
            This object.

        '''
        self.model = model
        return self

    def get_prediction_column(self):
        '''
        Fetch the name of the prediction column.

        Returns
        -------
        str
            The name of the prediction column.

        '''
        return self.__class__.prediction_col

    def get_orig_classes(self):
        '''
        Fetch the names of the original target classes.

        Returns
        -------
        list
            The names of the original target classes, if model_type
            is "classification".

        '''
        return self.target_orig_classes

    def get_int_values(self):
        '''
        Fetch the integers assigned to the target classes.

        Returns
        -------
        list
            The integers assigned to the target classes, if model_type
            is "classification".

        '''
        return self.target_int_values

    def get_target_name(self):
        '''
        Fetch the name of the target variable.

        Returns
        -------
        str
            The name of the target variable..

        '''
        return self.target

    def predict(self):
        '''
        Empty implementation.

        Returns
        -------
        None.

        '''
        pass

    def predict_probability(self):
        '''
        Empty implementation.

        Returns
        -------
        None.

        '''
        pass


    def calc_prediction(self, X):
        '''
        Calculate prediction for regression or classification models.
        Dispatch task to either calc_regression_prediction or
        calc_classification_prediction.

        Parameters
        ----------
        X : data frame
            The input data frame for calculating model predictions.

        Returns
        -------
        data frame
            The model predictions.

        '''
        # Any other model type is blocked when constructing the Scorer object.
        if self.model_type == 'regression':
            return self.calc_regression_prediction(X)
        elif self.model_type == 'classification':
            return self.calc_classification_prediction(X)


    def calc_regression_prediction(self, X):
        '''
        Calculate prediction for regression models.

        Parameters
        ----------
        X : data frame
            The input data frame for calculating model predictions.

        Returns
        -------
        prediction_df : data frame
            The model predictions, under a column of name
            BaseModelScorer.prediction_col.
        '''
        pred_col = self.__class__.prediction_col
        prediction_df = pd.DataFrame(columns=[pred_col])
        prediction_df[pred_col] = self.predict(X)
        return prediction_df


    def calc_classification_prediction(self, X):
        '''
        Calculate prediction for classification models.

        Parameters
        ----------
        X : data frame
            The input data frame for calculating model predictions.

        Returns
        -------
        prediction_df : data frame
            The model predictions, under a column of name
            BaseModelScorer.prediction_col. In addition,
            the class probabilities for each class are returned in
            additional columns, with names equal to the class names
            contained in the object variable target_orig_classes.
        '''
        pred_col = self.__class__.prediction_col
        pred_cols = self.get_orig_classes().copy()
        pred_cols.append(pred_col)

        prediction_df = pd.DataFrame(columns=pred_cols)
        prediction_df[pred_col] = self.predict(X)

        y_prob = self.predict_probability(X)
        prediction_df[self.get_orig_classes()] = pd.DataFrame(y_prob)
        # Avoid spurious numeric noise by rounding
        prediction_df[self.get_orig_classes()] = prediction_df[self.get_orig_classes()].round(4)

        return prediction_df


class SklearnScorer(BaseModelScorer):
    '''
    SKlearn implementation of the model Scorer
    '''
    # class variables ----------------------------------------------------------

    # object methods -----------------------------------------------------------

    def __init__(self, model_type, target=None,target_orig_classes=[], target_int_values=[]):
        '''
        Initialize object by storing the relevant information.
        Use super class with model_type="sklearn".

        Parameters
        ----------
        model_origin : str
            The source of the model. Currently only "sklearn" implemented.
        model_type : str
            One of "classification" or "regression".
        target : str, optional
            The name of the target variable. The default is None.
        target_orig_classes : list of str, optional
            The list of class names, if model_type is "classification".
            The default is [].
        target_int_values : list of int, optional
            The list of class integer values, if model_type is "classification".
            The default is [].

        Returns
        -------
        None.

        '''
        super(self.__class__, self).__init__('sklearn', model_type, target, target_orig_classes, target_int_values)

        return None


    def setup(self, model):
        '''
        Implementation of the setup function from the base scorer.
        Store the model. If this is a bytes object, cloud-pickle it into
        the original model object.
        Uses cloudpickle.

        Parameters
        ----------
        model : pipeline, bytes or str
            The input model.

        Returns
        -------
        Itself.
            Use the base scorer to store model.

        '''
        if type(model)==bytes:
            import cloudpickle
            model=cloudpickle.loads(model)
        return super(self.__class__, self).setup(model)


    def predict(self, X):
        '''
        Apply model to data frame and return prediction array.

        Parameters
        ----------
        X : data frame
            The input data frame for calculating model predictions.

        Returns
        -------
        numpy.ndarray
            The predicted values, in a monodimensional array.

        '''
        return self.model.predict(X)


    def predict_probability(self, X):
        '''
        Apply model to data frame and return predicted class probabilities.

        Parameters
        ----------
        X : data frame
            The input data frame for calculating model predictions.

        Returns
        -------
        numpy.ndarray
            The predicted class probability, in an array of as many columns
            as there are classes.

        '''
        return self.model.predict_proba(X)



# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
# Evaluator classes

class BaseModelEvaluator():
    '''
    A base model Evaluator class to calculate performance metrics and variable
    importance.
    '''

    # class variables --------------------------------------------------
    variable_column = 'variable'
    value_column = 'value'
    relative_value_column = 'relative_value'

    #: str: prefix for json schema description
    evaluator_parms_description = 'init parameters for Evaluator'
    varimp_parms_description = 'data for variable importance'
    performance_parms_description = 'data for performance'

    # class utility functions --------------------------------------------------
    @classmethod
    def choose_evaluator(cls, model_type, sensitivity_metric):
        '''
        Class method.
        Invoke the appropriate Evaluator sub-class depending on the model type.

        Parameters
        ----------
        model_type : str
            One of "classification" or "regression".
        sensitivity_metric : str
            One of "rmse","mape", "r2" for regression, or "accuracy","fscore"
            for classification.

        Returns
        -------
        Either a RegressionModelEvaluator or a ClassificationModelEvaluator,
        depending on the value of model_type.
        '''

        if model_type == 'regression':
            return RegressionModelEvaluator(sensitivity_metric)
        elif model_type == 'classification':
            return ClassificationModelEvaluator(sensitivity_metric)

    @classmethod
    def get_negative_metrics(cls):
        '''
        Class method.
        Negative metrics = the lower, the better.

        Returns
        -------
        list
            A list of the supported negative metrics.

        '''
        return ['rmse', 'mape']

    @classmethod
    def get_positive_metrics(cls):
        '''
        Class method.
        Positive metrics = the higher, the better.

        Returns
        -------
        list
            A list of the supported positive metrics.
        '''
        return ['accuracy', 'fscore', 'r2']

    # object methods -----------------------------------------------------------

    def __init__(self, sensitivity_metric):
        '''
        Initialize object by storing the relevant information.
        Save instance attributes passed in into special
        instance variable __init_attributes__.

        Parameters
        ----------
        sensitivity_metric : : str
            One of "rmse","mape", "r2" for regression, or "accuracy","fscore"
            for classification.

        Raises
        ------
        NotImplementedError
            If the sensitivity metric is not implemented.

        Returns
        -------
        None.

        '''
        self.sensitivity_metric = sensitivity_metric
        # save a copy of the instance attributes that were passed to construction
        self.__init_attributes__ = self.__dict__.copy()
        if sensitivity_metric in self.__class__.get_positive_metrics():
            self.metric_direction = 1
        elif sensitivity_metric in self.__class__.get_negative_metrics():
            self.metric_direction = -1
        else:
            raise NotImplementedError(str(self.__class__)+'.__init__: The selected '+\
                                      sensitivity_metric+' sensitivity_metric is not implemented yet')

        return None

    def init_attributes_to_json(self):
        '''
        Collect the instance attributes defined at construction,
        and turn them into a json string.
        Used within a constructor method.

        Returns
        -------
        string
            JSON string containing instance attributes needed to
            configure scorer.

        '''
        return json.dumps(self.__init_attributes__)


    def setup(self, model_scorer):
        '''
        Store the model Scorer object in this Evaluator object.

        Parameters
        ----------
        model_scorer : subclass of BaseModelScorer
            The model Scorer for this Evaluator.

        Returns
        -------
        An object of this class.
        '''
        self.scorer = model_scorer
        return self

    def calc_permutation_sensitivity(self, X, y, num_repeats=3):
        '''
        Wrapper function to decide which version of permutation sensitivity
        to use, depending on the number of rows of the input data frame.
        For now, this limit is hard-coded to 10,000 rows, sampled to 5,000.

        Parameters
        ----------
        X : data frame
            The input data frame.
        y : numpy array or series
            The reference prediction.
        num_repeats : int, optional
            The number of times to repeat the calculation for each variable.
            The default is 3.

        Returns
        -------
        variable_importance_df : data frame
            The variable importance result.

        '''
        # If there are more than 10000 rows, calculate variable importance
        # ...with sampling
        if X.shape[0]>10000:
            #what fraction to sample to generate 5000 rows
            frac = 5000/X.shape[0]
            variable_importance_df = self.calc_permutation_sensitivity_with_sampling(X, y, num_repeats, sample_frac=frac)
        else:
            variable_importance_df = self.calc_permutation_sensitivity_no_sampling(X, y, num_repeats)

        return variable_importance_df


    def calc_permutation_sensitivity_no_sampling(self, X, y,num_repeats):
        '''
        Model agnostic variable importance using sensitivity
        needs multiple calls to model scoring
        In this version, no row sampling is done.

        Parameters
        ----------
        X : data frame
            The input data frame.
        y : numpy array or series
            The reference prediction.
        num_repeats : int
            The number of times to repeat the calculation for each variable.

        Returns
        -------
        sens_df : data frame
            The variable importance result.

        '''
        predictors = X.columns.tolist()
        sens_df = pd.DataFrame(predictors, columns=[self.__class__.variable_column])
        for pp in predictors:
            X_shuffle_pp = X.copy(deep=True)

            sensitivity_rep = [0]*num_repeats
            for rep in range(num_repeats):
                # Reshuffle column pp
                X_shuffle_pp[pp] = X[pp].sample(frac=1,random_state=rep).values
                # Score new dataset
                y_pred_pp = pd.Series(self.scorer.predict(X_shuffle_pp))
                sensitivity_rep[rep] = self.calc_individual_performance_metric(
                    self.sensitivity_metric, y_pred_pp, y)
            sensitivity = sum(sensitivity_rep)/num_repeats

            sens_df.loc[sens_df[self.__class__.variable_column] == pp, self.__class__.value_column] = \
                sensitivity if self.metric_direction == -1 else 1-sensitivity

        sens_df.sort_values(by=self.__class__.value_column, axis=0, ascending=False, inplace=True)
        max_sens = max(sens_df[self.__class__.value_column])
        sens_df[self.__class__.relative_value_column] = sens_df[self.__class__.value_column]/max_sens

        return sens_df


    def calc_permutation_sensitivity_with_sampling(self, X, y,num_repeats,sample_frac=0.3,replace=False):
        '''
        Model agnostic variable importance using sensitivity
        needs multiple calls to model scoring
        In this version, row sampling is performed.

        Parameters
        ----------
        X : data frame
            The input data frame.
        y : numpy array or series
            The reference prediction.
        num_repeats : int
            The number of times to repeat the calculation for each variable.
        sample_frac: real, optional''
            The fraction of rows to sample at every iteration.
            The default is 0.3.
        replace : bool, optional
            Whether to perform sampling with replacement. The default is False.

        Returns
        -------
        sens_df : data frame
            The variable importance result.

        '''
        predictors = X.columns.tolist()
        sens_df = pd.DataFrame(predictors, columns=[self.__class__.variable_column])
        if not isinstance(y,pd.Series):
            y=pd.Series(y)
        for pp in predictors:
            X_shuffle_pp = X.copy(deep=True)
            # Reshuffle column pp once
            X_shuffle_pp[pp] = X[pp].sample(frac=1,random_state=num_repeats+10).values

            sensitivity_rep = [0]*num_repeats
            for rep in range(num_repeats):
                # Score sample of new dataset (take SAME sample of y)
                y_pred_pp = pd.Series(self.scorer.predict(X_shuffle_pp.sample(frac=sample_frac, replace=replace,random_state=rep)))
                sensitivity_rep[rep] = self.calc_individual_performance_metric(
                    self.sensitivity_metric, y_pred_pp, y.sample(frac=sample_frac, replace=replace,random_state=rep))
            sensitivity = sum(sensitivity_rep)/num_repeats

            sens_df.loc[sens_df[self.__class__.variable_column] == pp, self.__class__.value_column] = \
                sensitivity if self.metric_direction == -1 else 1-sensitivity

        sens_df.sort_values(by=self.__class__.value_column, axis=0, ascending=False, inplace=True)
        max_sens = max(sens_df[self.__class__.value_column])
        sens_df[self.__class__.relative_value_column] = sens_df[self.__class__.value_column]/max_sens

        return sens_df


    def process(self, X, y=None):
        '''
        Calculate prediction, variable importance and performance
        (if label is provided). These are stored as object variables.
        If X has over 10,000 rows, the variable importance uses sampling to
        5,000 rows.

        Parameters
        ----------
        X : data frame
            The input data frame.
        y : series, optional
            The true label. The default is None.

        Returns
        -------
        None.

        '''
        if y is not None:
            y.reset_index(drop=True, inplace=True)
        # Calculate prediction data table, and isolate prediction column
        # ...into y_pred for calculating performance
        self.prediction_df = self.scorer.calc_prediction(X)
        y_pred = self.prediction_df[self.scorer.get_prediction_column()]
        # Calculate variable importance using permutation sensitivity
        self.variable_importance_df = self.calc_permutation_sensitivity(X, y_pred)

        # Calculate model performance (if y is available)
        self.performance_df = self.calc_model_performance(X, y_pred, y)

        return None


    def calc_individual_performance_metric(self, metric, y, y_pred, round_to=3):
        '''
        Individual call to a single performance metric.
        Implemented in sub classes.

        Parameters
        ----------
        metric : str
            The name of the performance metric.
        y : series
            The ground truth (label).
        y_pred : series
            The prediction.
        round_to : int, optional
            How many significant figures to round to. The default is 3.

        Returns
        -------
        None.

        '''
        pass


    def calc_model_performance(self, X, y_pred, y=None):
        '''
        Overall performance matrix.
        Implemented in sub classes.

        Parameters
        ----------
        metric : str
            The name of the performance metric.
        y : series
            The ground truth (label).
        y_pred : series
            The prediction.

        Returns
        -------
        None.

        '''
        pass

    def pack_json(self):
        '''
        Implemented in sub classes.

        Returns
        -------
        None.

        '''
        pass


class ClassificationModelEvaluator(BaseModelEvaluator):
    '''
    A classification model Evaluator class to calculate performance metrics and variable
    importance.
    '''
    # class utility functions --------------------------------------------------
    @classmethod
    def unpack_json(cls, received_json):
        '''
        Class method.
        Take the model JSON object of a classification model
        and re-build the original
        model metadata and metadata-processing classes.

        Parameters
        ----------
        received_json : str or dict
            The JSON object.

        Returns
        -------
        received_num_worker : NumericBinsWorker
            The object for binning numeric variables.
        received_cat_worker : CategoricalFrequenciesWorker
            The object for binning categorical variables.
        received_predicted_classes_df : data frame
            The baseline predictions.
        received_class_probabilities_df : data frame
            The baseline class probabilities, binned.
        received_model_scorer : a subclass of BaseModelScorer
            The model Scorer object.
        received_model_evaluator : ClassificationModelEvaluator
            The model Evaluator object.
        received_variable_importance_df : data frame
            The baseline variable importance.
        received_performance_df : data frame
            The baseline performance (empty if no target present).

        '''
        # Unpack json and create metadata workers
        #If the json is not in dictionary form, turn it into one
        if type(received_json)== str:
            received_json = json.loads(received_json)
        # Class probabilities (which were binned) and associated numeric bins worker
        received_num_worker, received_class_probabilities_df = mdd.NumericBinsWorker.unpack_json(
            received_json, return_df=True)

        # Predicted classes (which were binned) and associated categorical bins worker
        received_cat_worker, received_predicted_classes_df = mdd.CategoricalFrequenciesWorker.unpack_json(
            received_json, return_df=True)

        # Model scorer
        received_model_scorer = BaseModelScorer.unpack_json(received_json)
        # Variable importance, performance and associated model evaluator
        received_model_evaluator_attributes = json.loads(
            received_json['model_evaluator_json'])
        received_model_evaluator = cls(**received_model_evaluator_attributes)
        received_variable_importance_df = pd.DataFrame(
            json.loads(received_json['variable_importance_json']))
        received_performance_df = pd.DataFrame(
            json.loads(received_json['performance_json']))

        return (received_num_worker, received_cat_worker, received_predicted_classes_df,
                received_class_probabilities_df,
                received_model_scorer,
                received_model_evaluator, received_variable_importance_df, received_performance_df)

    # object methods -----------------------------------------------------------

    def setup(self, model_scorer):
        '''
        Store the specified model Scorer object in this object.
        Create a target map from origial class names to integer values.

        Parameters
        ----------
        model_scorer : A subclass of BaseModelScorer
            The Scorer object.

        Returns
        -------
        self
            This object.

        '''
        self.scorer = model_scorer
        self.target_map = pd.DataFrame({'target': model_scorer.get_orig_classes(), 'class': model_scorer.get_int_values()})

        return self


    def calc_individual_performance_metric(self, metric, y, y_pred, round_to=3):
        '''
        Individual call to a single performance metric.
        Calls sklearn.metrics to calculate the desired performance metric.

        Parameters
        ----------
        metric : str
            The name of the performance metric. One of "accuracy" or "fscore"
            (f1 weighted).
        y : series
            The ground truth (label).
        y_pred : series
            The prediction.
        round_to : int, optional
            How many significant figures to round to. The default is 3.


        Returns
        -------
        real
            The calculated performance metric.

        '''
        from sklearn.metrics import accuracy_score, f1_score

        if metric == 'accuracy':
            performance_metric = accuracy_score(y, y_pred)
        elif metric == 'fscore':
            # Default average is binary but we would need to define an int target class (pos_label)
            #np.unique(y_pred) is there to avoid warnings in case not all labels are represented
            performance_metric = f1_score(y, y_pred, average='weighted', labels=np.unique(y_pred))
        else:
            performance_metric = np.nan

        return round(performance_metric, round_to)


    def calc_model_performance(self, X, y_pred, y=None):
        '''
        If y is not null, calculate a performance table including
        precision, recall, fscore, support for each class.
        Otherwise return an empty data frame.
        Calls sklearn.metrics to calculate the desired performance metrics.

        Parameters
        ----------
        X : data frame
            The input dataset.
        y_pred : series
            The prediction.
        y : series, optional
            The ground truth (label). The default is None.

        Returns
        -------
        performance_df : data frame
            The performance table.

        '''
        from sklearn.metrics import precision_recall_fscore_support

        if y is None:
            performance_df = pd.DataFrame(
                columns=['precision', 'recall', 'fscore', 'support', 'class', 'target_class'])
        else:
            accuracy = self.calc_individual_performance_metric('accuracy', y, y_pred)
            #np.unique(y_pred) is there to avoid warnings in case not all labels are represented
            prfs = precision_recall_fscore_support(y, y_pred, labels=np.unique(y_pred))

            performance_df = pd.DataFrame(np.column_stack(prfs), columns=[
                                          'precision', 'recall', 'fscore', 'support'])
            performance_df['accuracy'] = accuracy
            performance_df['class'] = performance_df.index
            # Add actual target values
            performance_df = performance_df.merge(self.target_map, on='class')
            performance_df = pd.melt(performance_df, id_vars=['target', 'class'])

        return performance_df

    def pack_json(self):
        '''
        Take all information from this object, and turn it into
        JSON.

        Returns
        -------
        complete_json : str
            The complete description of this object as a JSON string.

        '''
        # Name of column containing the prediction
        pred_col = self.scorer.get_prediction_column()
        y_pred = self.prediction_df[[pred_col]].astype(object)
        y_prob = self.prediction_df[self.scorer.get_orig_classes()]

        # Decode predictions to original classes
        target_decode=dict(zip(self.scorer.get_int_values(),self.scorer.get_orig_classes()))
        #only decode if all the elements in the series to decode are contained in the mapper
        if all(item in y_pred[pred_col].unique().tolist() for item in list(target_decode.keys())):
            y_pred[pred_col]=y_pred[pred_col].map(target_decode)

        # Bin prediction and turn to json
        num_binner_parms = {'auto_detect_numeric': True, 'include_nulls': True,
                            'binning_type': '01',
                            'n_bins_max': 20,
                            'name_of_null_level': 'missing',
                            'name_of_outlier_level': 'outlier',
                            'numeric': []}

        num_schema, num_json_body, num_json_df =\
            mdd.NumericBinsWorker.pack_json(num_binner_parms, y_prob,
                                            'numeric_binner_json', 'binned_probabilities_json')

        # Store binned prediction
        self.class_probabilities_df = pd.DataFrame(json.loads(num_json_df))

        # Bin classes and turn to json
        cat_binner_parms = {'auto_detect_categories': True, 'include_nulls': True,
                            'empty_as_null': True,
                            'name_of_null_level': 'missing',
                            'categories': []}
        cat_schema, cat_json_body, cat_json_df =\
            mdd.CategoricalFrequenciesWorker.pack_json(cat_binner_parms, y_pred,
                                                       'cat_binner_json', 'binned_classes_json')
        # Store binned prediction
        self.predicted_classes_df = pd.DataFrame(json.loads(cat_json_df))

        # Model scorer
        scorer_schema, scorer_json = self.scorer.pack_json('scorer_json')

        # Model evaluator, variable importance and performance
        model_evaluator_json = self.init_attributes_to_json()
        variable_importance_json = self.variable_importance_df.to_json()
        performance_json = self.performance_df.to_json()

        evaluator_json = {'model_evaluator_json': model_evaluator_json,
                          'variable_importance_json': variable_importance_json,
                          'performance_json': performance_json}

        # Create labels and mini schema
        evaluator_parms_description = self.__class__.evaluator_parms_description
        varimp_parms_description = self.__class__.varimp_parms_description
        performance_parms_description = self.__class__.performance_parms_description
        evaluator_schema = {evaluator_parms_description: 'evaluator_parms',
                            varimp_parms_description: 'varimp_parms',
                            performance_parms_description: 'performance_parms'}
        # Build json: the schema and then the json body for each metadata class
        json_schema = {**num_schema, **cat_schema, **scorer_schema, **evaluator_schema}
        complete_json = {**evaluator_json, **num_json_body, **cat_json_body, **scorer_json}

        json_schema = json.dumps(json_schema)  # turn into a json string
        json_schema = {'schema': json_schema}  # create a dictionary

        # Build json: the schema and then the json body for each metadata class
        complete_json = {**json_schema, **complete_json}
        complete_json = json.dumps(complete_json)

        return (complete_json)
# ------------------------------------------------------------------------------


class RegressionModelEvaluator(BaseModelEvaluator):
    '''
    A regression model Evaluator class to calculate performance metrics and variable
    importance.
    '''

    # class utility functions --------------------------------------------------
    @classmethod
    def unpack_json(cls, received_json):
        '''
        Class method.
        Take the model JSON object of a regression model and re-build the original
        model metadata and metadata-processing classes.

        Parameters
        ----------
        received_json : str or dict
            The JSON object containing the model metadata.

        Returns
        -------
        received_num_worker : NumericBinsWorker
            The object for binning numeric variables.
        received_binned_prediction_df : data frame
            The baseline prediction table, binned.
        received_model_scorer : a subclass of BaseModelScorer
            The model Scorer object.
        received_model_evaluator : RegressionModelEvaluator
            The model Evaluator object.
        received_variable_importance_df : data frame
            The baseline variable importance.
        received_performance_df : data frame
            The baseline performance (empty if no target present).

        '''
        # Unpack json and create metadata workers
        #If the json is not in dictionary form, turn it into one
        if type(received_json)== str:
            received_json = json.loads(received_json)
        # Prediction (which was binned) and associated numeric bins worker
        received_num_worker, received_binned_prediction_df = mdd.NumericBinsWorker.unpack_json(
            received_json, return_df=True)

        # Model scorer
        received_model_scorer = BaseModelScorer.unpack_json(received_json)
        # Variable importance, performance and associated model evaluator
        received_model_evaluator_attributes = json.loads(
            received_json['model_evaluator_json'])
        received_model_evaluator = cls(**received_model_evaluator_attributes)
        received_variable_importance_df = pd.DataFrame(
            json.loads(received_json['variable_importance_json']))
        received_performance_df = pd.DataFrame(
            json.loads(received_json['performance_json']))

        return (received_num_worker, received_binned_prediction_df, received_model_scorer,
                received_model_evaluator, received_variable_importance_df, received_performance_df)

    # object methods -----------------------------------------------------------

    def setup(self, model_scorer):
        '''
        Store the model Scorer object in this Evaluator object.
        Use the base class method.

        Parameters
        ----------
        model_scorer : subclass of BaseModelScorer
            The model Scorer for this Evaluator.

        Returns
        -------
        An object of this class.
        '''

        return super(self.__class__, self).setup(model_scorer)


    def calc_individual_performance_metric(self, metric, y, y_pred, round_to=3):
        '''
        Individual call to a single performance metric.
        Calls sklearn.metrics to calculate the desired performance metric.

        Parameters
        ----------
        metric : str
            The name of the performance metric. One of "rmse",
            "mape" or "r2".
        y : series
            The ground truth (label).
        y_pred : series
            The prediction.
        round_to : int, optional
            How many significant figures to round to. The default is 3.


        Returns
        -------
        real
            The calculated performance metric.
        '''

        from sklearn.metrics import mean_squared_error, \
            mean_absolute_percentage_error, r2_score

        if metric == 'rmse':
            performance_metric = mean_squared_error(y, y_pred, squared=False)
        elif metric == 'mape':
            performance_metric = mean_absolute_percentage_error(y, y_pred)
        elif metric == 'r2':
            performance_metric = r2_score(y, y_pred)
        else:
            performance_metric = np.nan

        return round(performance_metric, round_to)


    def calc_residual_normality(self, y_res):
        '''
        Use scipy.stats with alpha=0.05 to evaluate residual normality.

        Parameters
        ----------
        y_res : vector of real.
            The residuals.

        Returns
        -------
        residual_normality : str
            Either 'normal', 'not normal' or nan.
        p : real
            The p value.

        '''

        # check if residual distribution is normal
        from scipy.stats import normaltest
        alpha = 0.05
        k2, p = normaltest(y_res)
        if p < alpha:
            residual_normality = 'not normal'
        elif p >= alpha:
            residual_normality = 'normal'
        else:
            residual_normality = np.nan

        return (residual_normality, p)


    def calc_model_performance(self, X, y_pred, y=None):
        '''
        If y is not null, calculate a performance table including
        rmse, mape, r2, residual normality and its p value.
        Otherwise return an empty data frame.

        Parameters
        ----------
        X : data frame
            The input dataset.
        y_pred : series
            The prediction.
        y : series, optional
            The ground truth (label). The default is None.

        Returns
        -------
        performance_df : data frame
            The performance table.

        '''

        df_cols = ['rmse', 'mape', 'r2', 'res_norm', 'res_norm_p']
        df_types = [pd.Series(dtype='float'), pd.Series(dtype='float'), pd.Series(
            dtype='float'), pd.Series(dtype='str'), pd.Series(dtype='float')]
        cols_types = dict(zip(df_cols, df_types))

        if y is None:
            performance_df = pd.DataFrame(cols_types)
        else:
            y = y.astype('float64')
            y_res = y-y_pred
            y_res = y_res.astype('float64')

            rmse = self.calc_individual_performance_metric('rmse', y, y_pred)
            mape = self.calc_individual_performance_metric('mape', y, y_pred)
            r2 = self.calc_individual_performance_metric('r2', y, y_pred)
            res_norm, res_norm_p = self.calc_residual_normality(y_res)
            prfs = [rmse, mape, r2, res_norm, res_norm_p]
            performance_df = pd.DataFrame(np.column_stack(prfs), columns=df_cols)

        return performance_df

    def pack_json(self):
        '''
        Take all information from this object, and turn it into
        JSON.

        Returns
        -------
        complete_json : str
            The complete description of this object as a JSON string.

        '''

        # Name of column containing the prediction
        pred_col = self.scorer.get_prediction_column()
        y_pred = self.prediction_df[[pred_col]]

        # Bin prediction and turn to json
        num_binner_parms = {'auto_detect_numeric': True, 'include_nulls': True,
                            'binning_type': 'quantile',
                            'n_bins_max': 10,
                            'name_of_null_level': 'missing',
                            'name_of_outlier_level': 'outlier',
                            'numeric': []}

        num_schema, num_json_body, num_json_df =\
            mdd.NumericBinsWorker.pack_json(num_binner_parms, y_pred,
                                            'numeric_binner_json', 'binned_prediction_json')
        # Store binned prediction
        self.binned_prediction_df = pd.DataFrame(json.loads(num_json_df))

        # Model scorer
        scorer_schema, scorer_json = self.scorer.pack_json('scorer_json')

        # Model evaluator, variable importance and performance
        model_evaluator_json = self.init_attributes_to_json()
        variable_importance_json = self.variable_importance_df.to_json()
        performance_json = self.performance_df.to_json()

        evaluator_json = {'model_evaluator_json': model_evaluator_json,
                          'variable_importance_json': variable_importance_json,
                          'performance_json': performance_json}

        # Create labels and mini schema
        evaluator_parms_description = self.__class__.evaluator_parms_description
        varimp_parms_description = self.__class__.varimp_parms_description
        performance_parms_description = self.__class__.performance_parms_description
        evaluator_schema = {evaluator_parms_description: 'evaluator_parms',
                            varimp_parms_description: 'varimp_parms',
                            performance_parms_description: 'performance_parms'}

        # Build json: the schema and then the json body for each metadata class
        json_schema = {**num_schema, **scorer_schema, **evaluator_schema}
        complete_json = {**evaluator_json, **num_json_body, **scorer_json}

        json_schema = json.dumps(json_schema)  # turn into a json string
        json_schema = {'schema': json_schema}  # create a dictionary

        # Build json: the schema and then the json body for each metadata class
        complete_json = {**json_schema, **complete_json}
        complete_json = json.dumps(complete_json)

        return (complete_json)


# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------

# INTERNAL ENTRY POINTs


def init_scorer(score_mechanism, model, model_type, target=None,
                target_orig_classes=[], target_int_values=[]):
    '''
    Create a new Scorer object of the desired type.

    Parameters
    ----------
    score_mechanism : str
        The source of the model. Currently only "sklearn" implemented.
    model : bytes or sklearn pipeline
        The model object.
    model_type : str
        One of "classification" or "regression".
    target : str, optional
        The name of the target variable. The default is None.
    target_orig_classes : list of str, optional
        The list of class names, if model_type is "classification".
        The default is [].
    target_int_values : list of int, optional
        The list of class integer values, if model_type is "classification".
        The default is [].

    Returns
    -------
    scorer : Subclass of BaseModelScorer
        Currently returns an instance of SklearnScorer, or None if
        model_origin not "sklearn".

    '''
    scorer = BaseModelScorer.choose_scorer(score_mechanism, model_type, target,target_orig_classes, target_int_values)
    scorer.setup(model)

    return scorer


def init_evaluator(scorer, sensitivity_metric):
    '''
    Invoke the appropriate Evaluator sub-class depending on the model type.

    Parameters
    ----------
    scorer : Subclass of BaseModelScorer
        The model Scorer object to use.
    sensitivity_metric : str
        One of "rmse","mape", "r2" for regression, or "accuracy","fscore"
        for classification.

    Returns
    -------
    Either a RegressionModelEvaluator or a ClassificationModelEvaluator,
    depending on the value of the model_type stored within the Scorer
    object.
    '''
    model_type = scorer.model_type
    evaluator = BaseModelEvaluator.choose_evaluator(model_type, sensitivity_metric)
    evaluator.setup(scorer)

    return evaluator



def unpack_classification_model_metadata(complete_json, model):
    '''
    Take the model JSON string of a classification model and re-build the original
    model metadata and metadata-processing classes. Add the model to the Scorer
    object and add this to the Evaluator object.

    Parameters
    ----------
    complete_json : str or dict
        The JSON object containing the model metadata.
    model : bytes or sklearn pipeline
        The classification model object.

    Returns
    -------
    received_num_worker : NumericBinsWorker
        The object for binning numeric variables.
    received_cat_worker : CategoricalFrequenciesWorker
        The object for binning categorical variables.
    received_predicted_classes_df : data frame
        The baseline predictions.
    received_class_probabilities_df : data frame
        The baseline class probabilities, binned.
    received_model_scorer : a subclass of BaseModelScorer
        The model Scorer object.
    received_model_evaluator : RegressionModelEvaluator
        The model Evaluator object.
    received_variable_importance_df : data frame
        The baseline variable importance.
    received_performance_df : data frame
        The baseline performance (empty if no target present).

    '''
    received_num_worker, received_cat_worker, received_predicted_classes_df, \
        received_class_probabilities_df,\
        received_model_scorer,\
        received_model_evaluator, received_variable_importance_df, received_performance_df\
        = ClassificationModelEvaluator.unpack_json(complete_json)

    received_model_scorer.setup(model)
    received_model_evaluator.setup(received_model_scorer)

    return (received_num_worker, received_cat_worker,
            received_predicted_classes_df, received_class_probabilities_df,
            received_model_scorer,
            received_model_evaluator, received_variable_importance_df, received_performance_df)


def unpack_regression_model_metadata(complete_json, model):
    '''
    Take the model JSON string of a regression model and re-build the original
    model metadata and metadata-processing classes. Add the model to the Scorer
    object and add this to the Evaluator object.

    Parameters
    ----------
    complete_json : str or dict
        The JSON object containing the model metadata.
    model : bytes or sklearn pipeline
        The regression model object.

    Returns
    -------
    received_num_worker : NumericBinsWorker
        The object for binning numeric variables.
    received_binned_prediction_df : data frame
        The baseline prediction table, binned.
    received_model_scorer : a subclass of BaseModelScorer
        The model Scorer object.
    received_model_evaluator : RegressionModelEvaluator
        The model Evaluator object.
    received_variable_importance_df : data frame
        The baseline variable importance.
    received_performance_df : data frame
        The baseline performance (empty if no target present).

    '''
    received_num_worker, received_binned_prediction_df, received_model_scorer,\
        received_model_evaluator, received_variable_importance_df, received_performance_df\
        = RegressionModelEvaluator.unpack_json(complete_json)

    received_model_scorer.setup(model)
    received_model_evaluator.setup(received_model_scorer)

    return (received_num_worker, received_binned_prediction_df, received_model_scorer,
            received_model_evaluator, received_variable_importance_df, received_performance_df)


#
#
# --- USER ENTRY POINTs--------------------------------------------------------

def calculate_classification_model_metadata(model, df, target,encode_target,
                            target_orig_classes,target_int_values,scorer_type='sklearn'):
    '''
    Calculate initial model metadata and return JSON object. In addition, return
    initial variable importance, performance (if target present in input dataset),
    predictions and class probabilities tables.

    Parameters
    ----------
    model : bytes or sklearn pipeline
        The model to use for predictions.
    df : data frame
        The input dataset. This is supposed to contain only the predictor
        variables. It might also contain the target variable, but this will
        be removed before processing.
    target : str
        The name of classification target variable.
    encode_target : bool
        If True, we assume the target column has been encoded when training
        the model.
    target_orig_classes : list
        The names of the original target classes, before any encoding.
    target_int_values : TYPE
        The target integer values, in order.
    scorer_type : str, optional
        The type of Scorer object to initialize. Currently
        only "sklearn" (the default) is supported.

    Returns
    -------
    complete_json : str
        The model metadata.
    data frame
        variable importance.
    data frame
        performance. Empty if no target variable present.
    data frame
        predictions.
    data frame
        class probabilities. These are binned to 20 bins.

    '''

    # Prepare dataset
    predictors = list(df.columns)
    if target in predictors:
        predictors.remove(target)
        y = df[target]
    else:
        y = None
    X = df[predictors]

    # Create a target map: from original values to integers
    target_orig_classes=target_orig_classes.split(',')
    target_int_values=target_int_values.split(',')
    target_encode0=dict(zip(target_orig_classes,[int(x) for x in target_int_values]))
    #if we had encoded the target when generating the model, do it here
    if encode_target and y is not None:
        y=y.map(target_encode0)
        y=y.astype(int)
    # Initiate scorer and evaluator ----------------------------------------------------------------
    model_scorer = init_scorer(scorer_type, model, 'classification', target, target_orig_classes, target_int_values)
    model_evaluator = init_evaluator(model_scorer, 'fscore')
    # Calculate metrics ----------------------------------------------------------------------------
    model_evaluator.process(X, y)

    # Build json object ----------------------------------------------------------------------------
    complete_json = model_evaluator.pack_json()

    return (complete_json, model_evaluator.variable_importance_df,
            model_evaluator.performance_df, model_evaluator.predicted_classes_df,
            model_evaluator.class_probabilities_df)


def calculate_regression_model_metadata(model, df, target,scorer_type='sklearn'):
    '''
    Calculate initial model metadata and return JSON object. In addition, return
    initial variable importance, performance (if target present in input dataset),
    predictions tables.

    Parameters
    ----------
    model : bytes or sklearn pipeline
        The model to use for predictions.
    df : data frame
        The input dataset. This is supposed to contain only the predictor
        variables. It might also contain the target variable, but this will
        be removed before processing.
    target : str
        The name of classification target variable.
    scorer_type : str, optional
        The type of Scorer object to initialize. Currently
        only "sklearn" (the default) is supported.

    Returns
    -------
    complete_json : str
        The model metadata.
    data frame
        variable importance.
    data frame
        performance. Empty if no target variable present.
    data frame
        predictions. These are binned to 10 quantile bins.

    '''
    # Prepare dataset
    predictors = list(df.columns)
    if target in predictors:
        predictors.remove(target)
        y = df[target]
    else:
        y = None
    X = df[predictors]

    # Initiate scorer and evaluator -------------------------------------------
    model_scorer = init_scorer(scorer_type, model, 'regression', target)
    model_evaluator = init_evaluator(model_scorer, 'mape')
    # Calculate metrics -------------------------------------------------------
    model_evaluator.process(X, y)

    # Build json object -------------------------------------------------------
    complete_json = model_evaluator.pack_json()

    return (complete_json, model_evaluator.variable_importance_df,
            model_evaluator.performance_df, model_evaluator.binned_prediction_df)


# MAIN ------------------------------------------------------------------------
if __name__ == '__main__':
    pass
